package net.shrine.adapter

import cats.effect.{ContextShift, IO, Timer}
import ch.qos.logback.classic.Level
import fs2.Stream
import net.shrine.adapter.dao.QueryResultStatus
import net.shrine.adapter.i2b2Protocol.{ErrorResponse, ReadInstanceResultsResponse, ReadQueryInstancesResponse, ReadQueryResultResponse, ShrineResponse}
import net.shrine.config.{ConfigExtensions, ConfigSource}
import net.shrine.http4s.catsio.{ExecutionContexts, RepeatedIOTask}
import net.shrine.hub.mom.ShrineMomClient
import net.shrine.log.Loggable
import net.shrine.problem.{AbstractProblem, JsonProblemDigest, ProblemSources, RawProblem}
import net.shrine.protocol.version.{DateStamp, NodeKey, QueryId}
import net.shrine.protocol.i2b2.AuthenticationInfo
import net.shrine.protocol.version.v2.{UpdateCrcQueuedResult, UpdateCrcQueuedResultWithError, UpdateCrcQueuedResultWithProgress, ResultStatus}

import scala.concurrent.duration.FiniteDuration
import scala.util.control.NonFatal

/**
  * Poll for results from QUEUED queries and send those results to the hub when found.
  *
  * @since 1.26.5.1
 */

object QueuedQueriesPoller extends Loggable {

  val queuedQueryTimeToLive:FiniteDuration = ConfigSource.config.getFiniteDuration("shrine.adapter.queuedQueryTimeToLive")

  val queuedQueryPollInterval:FiniteDuration = ConfigSource.config.getFiniteDuration("shrine.adapter.queuedQueryPollInterval")

  val queuedQueryRestTimeBeforePolling:FiniteDuration = ConfigSource.config.getFiniteDuration("shrine.adapter.queuedQueryRestTimeBeforePolling")

  val queuedQueryCrcPollHttpCallTimeout:FiniteDuration = ConfigSource.config.getFiniteDuration("shrine.adapter.queuedQueryCrcPollHttpCallTimeout")

  val crcRunQueryTimeLimit:FiniteDuration = ConfigSource.config.getFiniteDuration("shrine.adapter.crcRunQueryTimeLimit")

  implicit val timer: Timer[IO] = IO.timer(ExecutionContexts.crcExecutionContext)
  implicit val concurrentEffect: ContextShift[IO] = IO.contextShift(ExecutionContexts.crcExecutionContext)

  private val scheduler: RepeatedIOTask = RepeatedIOTask.scheduleWithFixedDelay(
    initialDelay = queuedQueryRestTimeBeforePolling,
    interval = queuedQueryPollInterval,
    task = () => checkForQueuedQueries(),
    name = "QueuedQueriesPoller"
  )

  def restart(): IO[Unit] = scheduler.restartIO()

  def stop(): IO[Unit] = scheduler.stopIO()

  lazy val readQueryResultAdapter:QueuedQueryInterrogator = QueuedQueryInterrogator(ConfigSource.config)

  /**
    * @return IO[true] if any queued queries are found, false if not
    */
  def checkForQueuedQueries():IO[Boolean] = {

    //if any potential queued queries are found, return true
    info(s"Starting checkForQueuedQueries")

//Check the database for not completed, not running queries
    val queuedQueriesIO: IO[Seq[QueryResultStatus]] = readQueryResultAdapter.selectQueuedQueryIds

    // Filter out queries that were very recently QUEUED
    val oldEnoughIO: IO[Seq[QueryResultStatus]] = queuedQueriesIO.map{ queuedQueries =>
      debug(s"Found ${queuedQueries.size} queries. ${queuedQueries.mkString(", ")}")
      val now = System.currentTimeMillis()
      queuedQueries.filter{ queuedQuery =>
        //Wait at least polling time plus the CRC time. queuedQuery.timestamp.getTime can be the start time
        (now - queuedQuery.timestamp.getTime) > queuedQueryRestTimeBeforePolling.toMillis + crcRunQueryTimeLimit.toMillis
      }
    }
//Ask the CRC for an update
    oldEnoughIO.flatMap{ oldEnough: Seq[QueryResultStatus] =>
      val oldQueryIds = oldEnough.map(_.networkQueryId).distinct

      val oldEnoughAndDistinct: Seq[QueryResultStatus] = oldQueryIds.flatMap { id =>
        oldEnough.collectFirst { case q: QueryResultStatus if q.networkQueryId == id => q } }

      info(s"Found ${oldEnoughAndDistinct.size} queries old enough to check. ${oldEnoughAndDistinct.mkString(", ")}")

      val askCrcIOs: Seq[IO[Unit]] = oldEnoughAndDistinct.map{ queuedQuery:QueryResultStatus =>
        debug(s"Will ask the CRC about query ${queuedQuery.networkQueryId}")
        val auth = AuthenticationInfo.noPassword(queuedQuery.username,queuedQuery.domain)
        readQueryResultAdapter.askCrc(
          queuedQuery.networkQueryId,
          queuedQueryCrcPollHttpCallTimeout,
          auth
        ).flatMap{ interpretAndMaybeSendAnUpdate(queuedQuery,_) }.handleErrorWith{
          case NonFatal(x) =>
            //trap NonFatal exceptions, log, and move on to the next query
            //this will handle bad behavior from the CRC, just stumbling on the problematic query and continuing
            IO(error(s"Caught exception in QueuedQueriesPoller while checking on query ${queuedQuery.networkQueryId}",x))
        }
      }
      //todo SHRINE2020-1310
      //todo after the fs2 and cats IO upgrade, shuffle the list of CRC IOs (or do a sort inspired by exponential back-off)
      //todo then race the clock - stop querying just after the next polling interval starts.
      //todo then change the default polling interval back to 30 seconds - shrine.adapter.queuedQueryRestTimeBeforePolling
      val itOfStreams = askCrcIOs.map(Stream.eval)
      val seqResultStatus = itOfStreams.foldLeft[Stream[IO,Unit]](Stream.empty){ (soFar,next) =>
        soFar.append(next)
      }.compile.drain

      //this polling worked - sleep a bit then keep going
      seqResultStatus.map(_ => true)
    }.handleErrorWith{
      case NonFatal(x) =>
        error(s"Caught exception in QueuedQueriesPoller.checkForQueuedQueries",x)
        //something went wrong - sleep a bit then keep going so long as there is no fatal exception
        IO(true)
    }
  }

  def interpretAndMaybeSendAnUpdate(
                                     oldQueryResult:QueryResultStatus,
                                     crcResponse:Either[ErrorResponse, ShrineResponse]
                                   ):IO[Unit] = {
    val queryId = new QueryId(oldQueryResult.networkQueryId)

    debug(s"$queryId's response is $crcResponse")
    val maybeUpdateFromCrcIO: IO[Option[UpdateCrcQueuedResult]] = crcResponse.fold({ errorResponse: ErrorResponse =>
      readQueryResultAdapter.storeErrorFromShrine(oldQueryResult,errorResponse.problem).map( _ => { //Give the CRC one chance - See SHRINE-3327
        errorResponse.problem match {
          case qrna:QueryResultNotAvailable =>
            error(s"Not reporting $qrna to the hub. If it is recurring please set the status for this query to FINISHED in the shrine_query_history.QUERY_RESULT table")
            None       //don't report QueryResultNotAvailable SHRINE-3585
          case rp:RawProblem => Some(UpdateCrcQueuedResultWithError(
            queryId = queryId,
            adapterNodeKey = NodeKey.localNodeKey,
            problem = JsonProblemDigest(rp),
            status = ResultStatus.ErrorFromCrc,
            statusMessage = Option(errorResponse.errorMessage),
            adapterTime = DateStamp.now
          ))
          case _ => throw new IllegalStateException("Error response problems must be subclasses of RawProblem")
        }
      })
    },{ shrineResponse: ShrineResponse  =>

      debug(s"interpretAndMaybeSendAnUpdate shrineResponse is $shrineResponse")

      val maybeUpdateResult: Option[UpdateCrcQueuedResult] = shrineResponse match {
        case rqrr:ReadQueryResultResponse =>
          val updateResult = rqrr.singleNodeResult.createUpdateCrcQueuedResult(
            queryId,
            rqrr.queryId,
            readQueryResultAdapter.obfuscator.obfuscatorParameters
          )
          if(updateResult.status.isFinal) Some(updateResult)
          else None //only send an update if there's finally a result
        case rirr:ReadInstanceResultsResponse => // this is an intermediate state unless there's an error. If so store it.
          val updateResult = rirr.results.head.createUpdateCrcQueuedResult(
            queryId,
            rirr.results.head.resultId,
            readQueryResultAdapter.obfuscator.obfuscatorParameters
          )
          updateResult match {
            case updateResultWithError:UpdateCrcQueuedResultWithError =>
              //If this update is an error then store it.
              // A ReadInstanceResultsResponse - if in error - will likely not be stored
              readQueryResultAdapter.storeErrorFromShrine(oldQueryResult, updateResultWithError.problem)
              Some(updateResultWithError)
            case _ => None
          }
        case rqir:ReadQueryInstancesResponse => // this is an intermediate state unless there's an error. If so store it.
          val maybeUpdateResult = rqir.createUpdateResult(queryId)
          maybeUpdateResult.foreach {
            case updateResultWithError:UpdateCrcQueuedResultWithError => readQueryResultAdapter.storeErrorFromShrine(oldQueryResult, updateResultWithError.problem)
            case _ =>
          }
          maybeUpdateResult
        case _ => throw new IllegalArgumentException(s"Code does not handle a ${shrineResponse.getClass.getSimpleName} $shrineResponse")
      }
      IO(maybeUpdateResult)
    })

    val maybeUpdateIO: IO[Option[UpdateCrcQueuedResult]] = maybeUpdateFromCrcIO.map(maybeUpdateFromCrc => {
      maybeUpdateFromCrc.fold{
        maybeQueryTooOld(oldQueryResult)  //no update from the CRC, so check timeout
      }{
        //still in some flavor of QUEUED
        case progress: UpdateCrcQueuedResultWithProgress => maybeQueryTooOld(oldQueryResult).orElse(Some(progress))
        case update => Some(update)
      }
    })
    maybeUpdateIO.flatMap( maybeUpdate => maybeUpdate.fold(IO.unit)(sendUpdateResultMessageIO))
  }

  private def maybeQueryTooOld(oldQueryResult:QueryResultStatus):Option[UpdateCrcQueuedResult] = {
    val now = System.currentTimeMillis()
    if ((now - oldQueryResult.timestamp.getTime) > queuedQueryTimeToLive.toMillis ) {
      val queryId = new QueryId(oldQueryResult.networkQueryId)
      val problem = QueryQueuedTimeToLiveExceeded(queryId,queuedQueryTimeToLive)

      //store error to the database
      readQueryResultAdapter.storeErrorFromShrine(oldQueryResult,problem)

      Some(UpdateCrcQueuedResultWithError(
        queryId = queryId,
        adapterNodeKey = NodeKey.localNodeKey,
        problem = JsonProblemDigest(problem),
        status = ResultStatus.ErrorFromCrc,
        statusMessage = Option(problem.description),
        adapterTime = DateStamp.now
      ))
    }
    else None
  }

  private def sendUpdateResultMessageIO(updateResult:UpdateCrcQueuedResult): IO[Unit] = {
    ShrineMomClient.sendToHubIO(updateResult.queryId,updateResult,UpdateCrcQueuedResult,s" updateResult for ${updateResult.queryId} to the hub")
  }
}

//noinspection ScalaFileName
case class QueryQueuedTimeToLiveExceeded(queryId: QueryId, queryTimeToLive:FiniteDuration) extends AbstractProblem(ProblemSources.Adapter) {
  override def logLevel: Level = Level.DEBUG
  override val throwable: Option[Throwable] = None
  override val summary: String = s"Shrine's adapter has given up on this query after $queryTimeToLive"
  override val description = s"Shrine's adapter has polled the CRC for a result from query $queryId for $queryTimeToLive before giving up."
  override def detailsText: None.type = None
}
