package net.shrine.adapter.audit

import java.sql.SQLException
import cats.effect.IO

import javax.sql.DataSource
import com.typesafe.config.Config
import net.shrine.adapter.i2b2Protocol.{ErrorResponse, RunQueryRequest, RunQueryResponse, ShrineResponse}
import net.shrine.audit.{LongQueryId, QueryName, ShrineNodeId, Time, UserDomainName, UserName}
import net.shrine.log.Loggable
import net.shrine.slick.TestableDataSourceCreator
import net.shrine.config.{ConfigExtensions, ConfigSource}
import net.shrine.http4s.catsio.{ExecutionContexts, SimpleAsyncExecutor}
import net.shrine.protocol.version.v2.{RunQueryForResult, UpdateResult, UpdateResultWithCount}
import slick.jdbc.JdbcProfile

import scala.concurrent.duration.{Duration, DurationInt}
import scala.concurrent.{Await, Future, blocking}
import scala.language.postfixOps

/**
 * DB code for the Adapter audit metrics.
 *
 * @author david
 * @since 8/25/15
 */
case class AdapterAuditDb(schemaDef:AdapterAuditSchema,dataSource: DataSource) {
  import schemaDef._
  import jdbcProfile.api._
  import schemaDef.jdbcProfile.backend.DatabaseDef

  val database:DatabaseDef = Database.forDataSource(
    ds = dataSource,
    maxConnections = None,
    executor = SimpleAsyncExecutor(ExecutionContexts.databaseExecutionContext)
  )
  def createTables(): Unit = schemaDef.createTables(database)

  def dropTables(): Unit = schemaDef.dropTables(database)

  def dbRun[R](action: DBIOAction[R, NoStream, Nothing]):R = {
    val future: Future[R] = database.run(action)
    try {
      Await.result(future, AdapterAuditSchema.config.get("timeout", Duration.apply))
    } catch {
      case x: Throwable =>
        error("Caught ",x)
        throw x
    }
  }

  private val databaseContextShift = IO.contextShift(ExecutionContexts.databaseExecutionContext)

  private def runIO[R](dbio: DBIOAction[R, NoStream, _]): IO[R] = {
    IO.fromFuture(IO(blocking(database.run(dbio))))(databaseContextShift)
  }


  def insertQueryReceivedIO(runQueryForResult: RunQueryForResult):IO[Unit] = {
    debug(s"insertQueryReceived $runQueryForResult")

    insertQueryReceivedIO(QueryReceived.fromRunQueryForResult(runQueryForResult))
  }

  def insertQueryReceivedIO(queryReceived:QueryReceived):IO[Unit] = {
    runIO(allQueriesReceived += queryReceived).flatMap(_ => IO.unit)
  }

  //only used in test code
  def selectAllQueriesReceived:Seq[QueryReceived] = {
    dbRun(allQueriesReceived.result)
  }

  def insertExecutionStarted(runQueryRequest: RunQueryRequest):Unit = {
    debug(s"insertExecutionStarted $runQueryRequest")

    insertExecutionStarted(ExecutionStarted.fromRequest(runQueryRequest))
  }

  def insertExecutionStarted(executionStart:ExecutionStarted):Unit = {
    dbRun(allExecutionsStarted += executionStart)
  }

  def selectAllExecutionStarts:Seq[ExecutionStarted] = {
    dbRun(allExecutionsStarted.result)
  }

  def insertExecutionCompletedShrineResponse(request: RunQueryRequest,shrineResponse: Either[ErrorResponse, RunQueryResponse]): Unit = {
    debug(s"insertExecutionCompleted $shrineResponse for $request")

    shrineResponse.map(ExecutionCompleted.fromRequestResponse(request,_)).foreach(insertExecutionCompleted)
  }

  def insertExecutionCompleted(executionCompleted:ExecutionCompleted):Unit = {
    dbRun(allExecutionsCompleted += executionCompleted)
  }

  def selectAllExecutionCompletes:Seq[ExecutionCompleted] = {
    dbRun(allExecutionsCompleted.result)
  }
  def insertResultSent(networkQueryId: LongQueryId, shrineResponse:ShrineResponse):Unit = {
    debug(s"insertResultSent $shrineResponse for $networkQueryId")

    ResultSent.fromResponse(networkQueryId,shrineResponse).foreach(insertResultSent)
  }

  def insertResultSentIO(networkQueryId: LongQueryId, shrineResponse:ShrineResponse):IO[Unit] = {
    debug(s"insertResultSent $shrineResponse for $networkQueryId")
    ResultSent.fromResponse(networkQueryId,shrineResponse).fold(IO.unit)(insertResultSentIO)
  }

  def insertResultSentIO(updateResult:UpdateResult, queryName:String): IO[Unit] = {
    debug(s"insertResultSent $updateResult for $queryName")
    ResultSent.fromUpdateResult(updateResult, queryName).fold(IO.unit)(insertResultSentIO)
  }

  def insertResultSent(resultSent: ResultSent):Unit = {
    dbRun(allResultsSent += resultSent)
  }

  def insertResultSentIO(resultSent:ResultSent):IO[Unit] = {
    runIO(allResultsSent += resultSent).flatMap(_ => IO.unit)
  }

  def selectAllResultsSent:Seq[ResultSent] = {
    dbRun(allResultsSent.result)
  }
}

/**
 * Separate class to support schema generation without actually connecting to the database.
 *
 * @param jdbcProfile Database profile to use for the schema
 */
case class AdapterAuditSchema(jdbcProfile: JdbcProfile) extends Loggable {
  import jdbcProfile.api._

  def ddlForAllTables: jdbcProfile.DDL = {
    allQueriesReceived.schema ++
    allExecutionsStarted.schema ++
    allExecutionsCompleted.schema ++
    allResultsSent.schema
  }

  //to get the schema, use the REPL
  //println(AdapterAuditSchema.schema.ddlForAllTables.createStatements.mkString(";\n"))

  def createTables(database:Database): Unit = {
    try {
      val future = database.run(ddlForAllTables.create)
      Await.result(future,10 seconds)
    } catch {
      //I'd prefer to check and create schema only if absent. No way to do that with Oracle.
      case x:SQLException => info("Caught exception while creating tables. Recover by assuming the tables already exist.",x)
    }
  }

  def dropTables(database:Database): Unit = {
    val future = database.run(ddlForAllTables.drop)
    //Really wait forever for the cleanup
    Await.result(future,Duration.Inf)
  }

  class QueriesReceivedAuditTable(tag:Tag) extends Table[QueryReceived](tag,"queriesReceived") {
    def shrineNodeId = column[ShrineNodeId]("shrineNodeId")
    def userName = column[UserName]("userName")
    def networkQueryId = column[LongQueryId]("networkQueryId")
    def queryName = column[QueryName]("queryName")
    def timeQuerySent = column[Time]("timeQuerySent")
    private def timeQueryReceived = column[Time]("timeReceived")
    def userDomainName = column[Option[UserDomainName]]("userDomainName")

    def * = (shrineNodeId,userName,networkQueryId,queryName,timeQuerySent,timeQueryReceived,userDomainName) <> (QueryReceived.tupled,QueryReceived.unapply)
  }

  val allQueriesReceived = TableQuery[QueriesReceivedAuditTable]

  class ExecutionsStartedTable(tag:Tag) extends Table[ExecutionStarted](tag,"executionsStarted") {
    def networkQueryId = column[LongQueryId]("networkQueryId")
    def queryName = column[QueryName]("queryName")
    private def timeExecutionStarts = column[Time]("timeExecutionStarted")

    def * = (networkQueryId,queryName,timeExecutionStarts) <> (ExecutionStarted.tupled,ExecutionStarted.unapply)
  }
  
  val allExecutionsStarted = TableQuery[ExecutionsStartedTable]

  class ExecutionsCompletedTable(tag:Tag) extends Table[ExecutionCompleted](tag,"executionsCompleted") {
    def networkQueryId = column[LongQueryId]("networkQueryId")
    def replyId = column[Long]("replyId")
    def queryName = column[QueryName]("queryName")
    private def timeExecutionCompletes = column[Time]("timeExecutionCompleted")

    def * = (networkQueryId,replyId,queryName,timeExecutionCompletes) <> (ExecutionCompleted.tupled,ExecutionCompleted.unapply)
  }

  val allExecutionsCompleted = TableQuery[ExecutionsCompletedTable]

  class ResultsSentTable(tag:Tag) extends Table[ResultSent](tag,"resultsSent") {
    def networkQueryId = column[LongQueryId]("networkQueryId")
    def replyId = column[Long]("replyId")
    def queryName = column[QueryName]("queryName")
    def timeResultsSent = column[Time]("timeResultsSent")

    def * = (networkQueryId,replyId,queryName,timeResultsSent) <> (ResultSent.tupled,ResultSent.unapply)
  }
  val allResultsSent = TableQuery[ResultsSentTable]
}

object AdapterAuditSchema {

  val config:Config = ConfigSource.config.getConfig("shrine.adapter.audit.database")

  val slickProfile:JdbcProfile = TestableDataSourceCreator.slickDriver(config)

  val schema: AdapterAuditSchema = AdapterAuditSchema(slickProfile)
}

object AdapterAuditDb {

  val dataSource:DataSource = TestableDataSourceCreator.dataSource(AdapterAuditSchema.config)

  val db: AdapterAuditDb = AdapterAuditDb(AdapterAuditSchema.schema,dataSource)

  val createTablesOnStart: Boolean = AdapterAuditSchema.config.getBoolean("createTablesOnStart")
  if(createTablesOnStart) AdapterAuditDb.db.createTables()

}

case class QueryReceived(
                          shrineNodeId:ShrineNodeId,
                          userName:UserName,
                          networkQueryId:LongQueryId,
                          queryName:QueryName,
                          timeQuerySent:Time,
                          timeQueryReceived:Time,
                          userDomainName:Option[UserDomainName]
                        )

object QueryReceived extends ((
  ShrineNodeId,
    UserName,
    LongQueryId,
    QueryName,
    Time,
    Time,
    Option[UserDomainName]
  ) => QueryReceived) with Loggable {

  def fromRunQueryForResult(runQueryForResult: RunQueryForResult):QueryReceived = {

    QueryReceived(
      shrineNodeId = runQueryForResult.node.name.underlying,
      userName = runQueryForResult.researcher.userName.underlying,
      networkQueryId = runQueryForResult.query.id.underlying,
      queryName = runQueryForResult.query.queryName,
      timeQuerySent = runQueryForResult.query.versionInfo.createDate.underlying,
      timeQueryReceived = System.currentTimeMillis(),
      userDomainName = Some(runQueryForResult.researcher.userDomainName.underlying)
    )
  }
}

case class ExecutionStarted(
                             networkQueryId:LongQueryId,
                             queryName:QueryName,
                             timeExecutionStarted:Time
                             )

object ExecutionStarted extends ((
  LongQueryId,
    QueryName,
    Time
  ) => ExecutionStarted){
  def fromRequest(rqr:RunQueryRequest): ExecutionStarted = {

    ExecutionStarted(rqr.networkQueryId,
      rqr.queryDefinition.name,
      System.currentTimeMillis())
  }
}

case class ExecutionCompleted(
                               networkQueryId:LongQueryId,
                               replyId:Long,
                               queryName:QueryName,
                               timeExecutionCompleted:Time
                             )

object ExecutionCompleted extends ((
    LongQueryId,
    Long,
    QueryName,
    Time
  ) => ExecutionCompleted){
  def fromRequestResponse(request: RunQueryRequest,rqr:RunQueryResponse): ExecutionCompleted = {
    ExecutionCompleted(
        request.networkQueryId,
        rqr.queryId,
        rqr.queryName,
        System.currentTimeMillis())
  }
}

case class ResultSent(
                       networkQueryId:LongQueryId,
                       responseId:Long,
                       queryName:QueryName,
                       timeResultSent:Time
                       )

object ResultSent extends ((
  LongQueryId,
    Long,
    QueryName,
    Time
  ) => ResultSent){
  def fromResponse(networkQueryId:LongQueryId, shrineResponse:ShrineResponse): Option[ResultSent] = {

    shrineResponse match {
      case rqr:RunQueryResponse => Some(ResultSent(
        networkQueryId,
        rqr.queryId,
        rqr.queryName,
        System.currentTimeMillis()))
      case _ => None
    }
  }

  def fromUpdateResult(updateResult:UpdateResult, queryName:String):Option[ResultSent] = {
    updateResult match {
      case urwcr:UpdateResultWithCount => Option( ResultSent(
        urwcr.result.queryId.underlying,
        urwcr.result.crcQueryInstanceId.get,
        queryName,
        System.currentTimeMillis()
      ))
      case _ => None
    }
  }
}

