package net.shrine.aim3

import cats.effect.IO
import com.typesafe.config.Config
import net.shrine.config.ConfigSource
import net.shrine.http4s.catsio.ExecutionContexts
import net.shrine.log.Log
import org.http4s.{HttpRoutes, Request, Response}
import org.http4s.dsl.impl.{->, /, Auth}
import org.http4s.dsl.io.{GET, InternalServerError, NotFound, Ok, POST, Root, http4sInternalServerErrorSyntax, http4sNotFoundSyntax, http4sOkSyntax}

import scala.util.control.NonFatal

case class Aim3Service(pmAuthorizer: PmAuthorizer = PmAuthorizerComponent.create) extends Auth {

  val aim3ServiceInfo: String =
    """
      |The SHRINE aim3 service.
      |
      |This API retrieves data from the SHRINE adapter's database for the aim3 plugins.
    """.stripMargin

  val service: HttpRoutes[IO] = HttpRoutes.of[IO] {
    case GET -> Root / "ping" => Ok("pong")
    case req @ POST -> Root / "xml"  => xmlRequestService(req)
    case GET -> Root => Ok(aim3ServiceInfo)
    case x => NotFound(s"The aim3 service does not respond to $x")
  }

  val config: Config = ConfigSource.config
  val shrineConfig: Config = config.getConfig("shrine")
  val aim3XmlRequestHandler:Aim3XmlRequestHandler = Aim3XmlRequestHandler.create(pmAuthorizer)

  //todo for the future json service do basic auth
  def xmlRequestService(request: Request[IO]): IO[Response[IO]] = {

    val httpClientContextShift = IO.contextShift(ExecutionContexts.httpClientExecutionContext)
    httpClientContextShift.evalOn(ExecutionContexts.httpClientExecutionContext) {
      request.as[String].flatMap[Response[IO]] { bodyText: String =>
        Log.debug(s"Received aim 3 request: $bodyText")

        aim3XmlRequestHandler.handleRequest(bodyText).fold({
          case NonFatal(x) =>
            Log.error(s"Exception while filling aim3 request", x)
            InternalServerError(x.getMessage)
        }, { aim3ResponseIO =>
          aim3ResponseIO.map(responseBody => {
            Log.debug(s"Replying to aim 3 request with: $responseBody")
            responseBody.toI2b2String
          }).flatMap(Ok(_))
        }) //Try
      } //IO
    } //evalOn
  }
}
