package net.shrine.aim3

import cats.effect.IO
import com.typesafe.config.Config
import net.shrine.adapter.i2b2Protocol.{HiveCredentials, ReadPreviousQueriesResponse, ShrineResponse}
import net.shrine.authentication.pm.User
import net.shrine.config.{ConfigExtensions, ConfigSource}
import net.shrine.protocol.i2b2.ShrineRequest

import scala.util.Try
import scala.xml.NodeSeq

/**
 * Copied and bent from I2b2AdminService
 */
final case class Aim3XmlRequestHandler(
                                        i2b2AdminDao: I2b2AdminDao,
                                        pmAuthorizerComponent: PmAuthorizer
                                      )  {

  def handleRequest(i2b2XmlRequestBody:String): Try[IO[ShrineResponse]] = {
    import net.shrine.xml.StringEnrichments.HasStringEnrichments
    val i2b2XmlTry: Try[NodeSeq] = i2b2XmlRequestBody.tryToXml
    i2b2XmlTry.flatMap {
      case i2b2Xml@previousQueriesXml if ReadI2b2AdminPreviousQueriesRequest.isReadI2b2AdminPreviousQueriesRequest(previousQueriesXml) =>
        ReadI2b2AdminPreviousQueriesRequest.fromI2b2(Set.empty)(i2b2Xml).map { pqr: ReadI2b2AdminPreviousQueriesRequest =>
          checkWithPmAndThen(pqr) { pqr =>
            val queriesIO = i2b2AdminDao.findQueriesByRequestIO(pqr)
            queriesIO.map(q => ReadPreviousQueriesResponse(q.map(_.toQueryMaster(_.i2b2MasterQueryId))))
          }
        }
      case i2b2Xml@usersXml if ReadI2b2AdminQueryingUsersRequest.isI2b2AdminQueryingUsersRequest(usersXml) =>
        ReadI2b2AdminQueryingUsersRequest.fromI2b2(Set.empty)(i2b2Xml).map { ur =>
          checkWithPmAndThen(ur) { ur =>
            val usersIO = i2b2AdminDao.findQueryingUsersByProjectIdIO(ur.projectIdToQueryFor)
            usersIO.map(u => ReadI2b2AdminQueryingUsersResponse(u))
          }
        }
      case _ => throw UnknownMessageBodyElementException(i2b2XmlRequestBody)
    }
  }

  def checkWithPmAndThen[Req <: ShrineRequest](request: Req)(f: Req => IO[ShrineResponse]): IO[ShrineResponse] = {
    //todo change to/unify with AuthStrategy or the equivalent
    val authorized: AuthorizationStatus = pmAuthorizerComponent.authorize(request.projectId, Set(User.Roles.Manager), request.authn)
    authorized match {
      case PmAuthorized(_) => f(request)
      case na: PmNotAuthorized => IO(na.toErrorResponse)
    }
  }
}

object Aim3XmlRequestHandler {
  def create(pmAuthorizerComponent: PmAuthorizer): Aim3XmlRequestHandler = {
    val config: Config = ConfigSource.config
    val shrineConfig: Config = config.getConfig("shrine")


    val crcHiveCredentials: HiveCredentials = shrineConfig.getConfigured("hiveCredentials", HiveCredentials(_, HiveCredentials.CRC))
    val i2b2AdminDao: I2b2AdminDao = new I2b2AdminDao(crcHiveCredentials.projectId)

    Aim3XmlRequestHandler(i2b2AdminDao,  pmAuthorizerComponent)
  }
}

case class UnknownMessageBodyElementException(xmlString:String) extends Exception(s"Message body sub element not recognized in: $xmlString")


