package net.shrine.aim3

import ch.qos.logback.classic.Level
import com.typesafe.config.Config
import net.shrine.adapter.i2b2Protocol.ErrorResponse
import net.shrine.authentication.pm.{GetUserConfigurationRequest, User}
import net.shrine.config.{ConfigExtensions, ConfigSource}
import net.shrine.http4s.client.legacy.{EndpointConfig, HttpResponse, Poster}
import net.shrine.log.Loggable
import net.shrine.problem._
import net.shrine.protocol.i2b2.AuthenticationInfo

import scala.util.control.NonFatal
import scala.util.{Failure, Success, Try}

/**
 * @author clint
 * @since Apr 5, 2013
 */
trait PmAuthorizer extends Loggable {

  def pmPoster:Poster

  private def parsePmResult(authn: AuthenticationInfo)(httpResponse: HttpResponse): Try[Either[ErrorResponse, User]] = {
    User.fromI2b2(httpResponse.body).map(Right(_)).recoverWith {
      case NonFatal(e) =>
        info(s"Couldn't extract a User from '$httpResponse'",e)
        Try(Left(ErrorResponse.fromI2b2(httpResponse.body)))
    }.recover {
      case NonFatal(e) =>
        val problem = CouldNotInterpretResponseFromPmCell(pmPoster.url,authn,httpResponse,e)
        Left(ErrorResponse(problem))
    }
  }

  def authorize(projectId: String, neededRoles: Set[String], authn: AuthenticationInfo): AuthorizationStatus = {

    val request: GetUserConfigurationRequest = GetUserConfigurationRequest(authn)

    val responseAttempt: Try[HttpResponse] = Try {
      debug(s"Authorizing with PM cell at ${pmPoster.url}")
      pmPoster.post(request.toI2b2String)
    }

    val authStatusAttempt: Try[AuthorizationStatus with Product with Serializable] = responseAttempt.flatMap(parsePmResult(authn)).map {
      case Right(user) =>
        val managerUserOption = for {
          roles <- user.rolesByProject.get(projectId)
          if neededRoles.forall(roles.contains)
        } yield user

        managerUserOption.map(PmAuthorized).getOrElse {
          PmNotAuthorized(MissingRequiredRoles(projectId,neededRoles,authn))
        }
      case Left(errorResponse) =>
        //todo remove when ErrorResponse gets its message
        info(s"ErrorResponse message '${errorResponse.errorMessage}' may not have carried through to the NotAuthorized object")
        PmNotAuthorized(errorResponse.problem)
    }

    authStatusAttempt match {
      case Success(s) => s
      case Failure(x) => PmNotAuthorized(CouldNotReachPmCell(pmPoster.url,authn,x))
    }
  }
}

//todo move to Aim3 as part of SHRINE-3541, and combine with the PmAuthorizer trait
case class PmAuthorizerComponent(pmPoster:Poster) extends PmAuthorizer

object PmAuthorizerComponent {
  def create: PmAuthorizerComponent = {
    val config: Config = ConfigSource.config
    val shrineConfig: Config = config.getConfig("shrine")

    val pmEndpoint: EndpointConfig = shrineConfig.getConfigured("pmEndpoint", EndpointConfig(_))
    val pmPoster: Poster = Poster(pmEndpoint)
    PmAuthorizerComponent(pmPoster)
  }
}

sealed trait AuthorizationStatus

case class PmAuthorized(user: User) extends AuthorizationStatus

case class PmNotAuthorized(problem: Problem) extends AuthorizationStatus {
  def toErrorResponse: ErrorResponse = ErrorResponse(problem.summary,problem)
}

case class MissingRequiredRoles(projectId: String, neededRoles: Set[String], authn: AuthenticationInfo) extends AbstractProblem(ProblemSources.Qep) {
  override def logLevel: Level = Level.INFO
  override val summary: String = s"User ${authn.domain}:${authn.username} is missing roles in project '$projectId'"

  override val description:String = s"User ${authn.domain}:${authn.username} does not have all the needed roles: ${neededRoles.map("'" + _ + "'").mkString(", ")} in the project '$projectId'"
}

case class CouldNotReachPmCell(pmUrl:String,authn: AuthenticationInfo,x:Throwable) extends AbstractProblem(ProblemSources.Qep) {
  override def logLevel: Level = Level.WARN
  override val throwable: Option[Throwable] = Some(x)
  override val summary: String = s"Could not reach PM cell."
  override val description:String = s"Shrine encountered ${throwable.get} while attempting to reach the PM cell at $pmUrl for ${authn.domain}:${authn.username}."
}

case class CouldNotInterpretResponseFromPmCell(pmUrl:String,authn: AuthenticationInfo,httpResponse: HttpResponse,x:Throwable) extends AbstractProblem(ProblemSources.Qep) {
  override def logLevel: Level = Level.ERROR
  override val throwable: Option[Throwable] = Some(x)
  override def summary: String = s"Could not interpret response from PM cell."

  override def description: String = s"Shrine could not interpret the response from the PM cell at $pmUrl for ${authn.domain}:${authn.username}: due to ${throwable.get}"
  override val detailsText: Option[String] = Some(s"Response is $httpResponse")
}