package net.shrine.aim3

import cats.data.OptionT
import cats.effect.{ConcurrentEffect, ContextShift, IO}
import com.typesafe.config.Config
import net.shrine.adapter.{AdapterTestHelpers, i2b2Protocol}
import net.shrine.adapter.dao.{AdapterQueryHistoryDb, Category, SortOrder, Strategy, UserNameQueryOptions}
import net.shrine.adapter.i2b2Protocol.query.{I2b2QueryDefinition, Term}
import net.shrine.adapter.i2b2Protocol.{QueryMaster, ReadPreviousQueriesResponse}
import net.shrine.authentication.pm.User
import net.shrine.xml.XmlDateHelper
import org.junit.{After, Before}

import scala.concurrent.duration.FiniteDuration
import net.shrine.http4s.client.legacy.{EndpointConfig, Poster}
import net.shrine.protocol.i2b2.{AuthenticationInfo, Credential}
import net.shrine.config.{ConfigExtensions, ConfigSource}
import org.http4s.{Method, Request, Response, Uri}
import org.http4s.dsl.io.Ok
import org.junit.Test
import org.scalatest.TestSuite

import scala.concurrent.ExecutionContext.Implicits.global

class Aim3ServiceTest extends TestSuite {

  //Aim3ServiceTest //first class object to load test data

  def aim3Service: Aim3Service = {
    implicit val contextShift: ContextShift[IO] = IO.contextShift(global)
    implicit val concurrentEffect: ConcurrentEffect[IO] = IO.ioConcurrentEffect(contextShift)

    val aim3Service = Aim3Service(AgreeablePmAuthorizer.create)
    aim3Service
  }

  def extractResponse(request: Request[IO]): Response[IO] = {

    val responseOptionIo: OptionT[IO, Response[IO]] = Aim3Service(AgreeablePmAuthorizer.create).service.run(request)

    responseOptionIo.map { r: Response[IO] => r }.fold {
      fail("No response from service")
    } { r: Response[IO] => r
    }.unsafeRunSync()
  }

  @Test
  def testPing():Unit = {
    val pingRequest: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString(s"/ping"))

    val response: Response[IO] = extractResponse(pingRequest)

    assertResult(Ok)(response.status)
    val entityText = response.as[String].unsafeRunSync()
    assertResult("pong")(entityText)
  }

  @Test
  def testReadI2b2AdminQueryingUsers():Unit = {
    val requestInstance = {
      ReadI2b2AdminQueryingUsersRequest(
        projectId = AdapterTestHelpers.projectId,
        waitTime = AdapterTestHelpers.waitTime,
        authn = AdapterTestHelpers.authn,
        projectIdToQueryFor = AdapterTestHelpers.projectId
      )
    }

    val request: Request[IO] = Request(method = Method.POST, uri = Uri.unsafeFromString(s"/xml")).withEntity(requestInstance.toI2b2String)

    val response: Response[IO] = extractResponse(request)
    assertResult(Ok)(response.status)

    val rpqResponseString = response.as[String]
    rpqResponseString.unsafeRunSync()

    val rpqResponse:ReadI2b2AdminQueryingUsersResponse = response.as[String].map(ReadI2b2AdminQueryingUsersResponse.fromI2b2).unsafeRunSync()
    val users: Seq[I2b2AdminUserWithRole] = rpqResponse.users
    val expected = Seq(I2b2AdminUserWithRole("Demo","a-different-user","USER"), I2b2AdminUserWithRole("Demo","some-other-user","USER"))

    assertResult(expected)(users)

  }

  @Test
  def testReadI2b2AdminPreviousQueriesAll():Unit = {
    val requestInstance = requestPreviousQueries
    val request: Request[IO] = Request(method = Method.POST, uri = Uri.unsafeFromString(s"/xml")).withEntity(requestInstance.toI2b2String)

    val response: Response[IO] = extractResponse(request)
    assertResult(Ok)(response.status)

    val rpqResponse:ReadPreviousQueriesResponse = response.as[String].map(ReadPreviousQueriesResponse.fromI2b2).unsafeRunSync()
    val queryMasterIds = rpqResponse.queryMasters.map(_.networkQueryId)
    assertResult(Seq(AdapterTestHelpers.networkQueryId4,AdapterTestHelpers.networkQueryId3,AdapterTestHelpers.networkQueryId2, AdapterTestHelpers.networkQueryId1))(queryMasterIds)
  }

  @Test
  def testReadPreviousQueriesExcludeUser(): Unit = {

    val requestInstance = requestPreviousQueries.copy(username = UserNameQueryOptions.Except(AdapterTestHelpers.authn2.username))
    val request: Request[IO] = Request(method = Method.POST, uri = Uri.unsafeFromString(s"/xml")).withEntity(requestInstance.toI2b2String)

    val response: Response[IO] = extractResponse(request)

    assertResult(Ok)(response.status)

    val rpqResponse:ReadPreviousQueriesResponse = response.as[String].map(ReadPreviousQueriesResponse.fromI2b2).unsafeRunSync()
    assertResult(Seq(AdapterTestHelpers.networkQueryId2, AdapterTestHelpers.networkQueryId1))(rpqResponse.queryMasters.map(_.networkQueryId))
  }

  //i2b2 aim3 case 4 and 5 - find all flagged queries regardless of user
  @Test
  def testReadPreviousQueriesOnlyFlagged(): Unit = {
    val requestInstance = requestPreviousQueries.copy(categoryToSearchWithin = Category.Flagged)

    val request: Request[IO] = Request(method = Method.POST, uri = Uri.unsafeFromString(s"/xml")).withEntity(requestInstance.toI2b2String)

    val response: Response[IO] = extractResponse(request)

    assertResult(Ok)(response.status)

    val rpqResponse:ReadPreviousQueriesResponse = response.as[String].map(ReadPreviousQueriesResponse.fromI2b2).unsafeRunSync()
    assertResult(Seq(AdapterTestHelpers.networkQueryId4, AdapterTestHelpers.networkQueryId1))(rpqResponse.queryMasters.map(_.networkQueryId))
  }

  @Test
  def testReadPreviousQueriesOnlyFlaggedExcludingUser(): Unit = {
    val requestInstance = requestPreviousQueries.copy(
      username = UserNameQueryOptions.Except(AdapterTestHelpers.authn.username),

      categoryToSearchWithin = Category.Flagged
    )
    val request: Request[IO] = Request(method = Method.POST, uri = Uri.unsafeFromString(s"/xml")).withEntity(requestInstance.toI2b2String)

    val response: Response[IO] = extractResponse(request)
    assertResult(Ok)(response.status)

    val rpqResponse:ReadPreviousQueriesResponse = response.as[String].map(ReadPreviousQueriesResponse.fromI2b2).unsafeRunSync()
    assertResult(Seq(AdapterTestHelpers.networkQueryId4))(rpqResponse.queryMasters.map(_.networkQueryId))
  }

  @Test
  def testReadPreviousQueriesWithSearchStringAndFlagged(): Unit = {
    val requestInstance: ReadI2b2AdminPreviousQueriesRequest = requestPreviousQueries.copy(
      searchString = AdapterTestHelpers.queryName1,
      categoryToSearchWithin = Category.Flagged
    )
    val request: Request[IO] = Request(method = Method.POST, uri = Uri.unsafeFromString(s"/xml")).withEntity(requestInstance.toI2b2String)

    val response: Response[IO] = extractResponse(request)
    assertResult(Ok)(response.status)

    val rpqResponse:ReadPreviousQueriesResponse = response.as[String].map(ReadPreviousQueriesResponse.fromI2b2).unsafeRunSync()
    assertResult(Seq(AdapterTestHelpers.networkQueryId1))(rpqResponse.queryMasters.map(_.networkQueryId))
  }

  @Test
  def testReadPreviousQueriesNoMatch(): Unit = {
    val requestInstance: ReadI2b2AdminPreviousQueriesRequest = requestPreviousQueries.copy(
      searchString = "some-query-name-that-doesn't-exist",
    )
    val request: Request[IO] = Request(method = Method.POST, uri = Uri.unsafeFromString(s"/xml")).withEntity(requestInstance.toI2b2String)

    val response: Response[IO] = extractResponse(request)
    assertResult(Ok)(response.status)

    val rpqResponse:ReadPreviousQueriesResponse = response.as[String].map(ReadPreviousQueriesResponse.fromI2b2).unsafeRunSync()
    assertResult(Seq.empty)(rpqResponse.queryMasters.map(_.networkQueryId))
  }

  @Test
  def testReadI2b2AdminPreviousQueriesDescending():Unit = {
    val requestInstance = requestPreviousQueries.copy(sortOrder = SortOrder.Descending)
    val request: Request[IO] = Request(method = Method.POST, uri = Uri.unsafeFromString(s"/xml")).withEntity(requestInstance.toI2b2String)

    val response: Response[IO] = extractResponse(request)
    assertResult(Ok)(response.status)

    val rpqResponse:ReadPreviousQueriesResponse = response.as[String].map(ReadPreviousQueriesResponse.fromI2b2).unsafeRunSync()
    val queryMasterIds = rpqResponse.queryMasters.map(_.networkQueryId)
    assertResult(Seq(AdapterTestHelpers.networkQueryId4,AdapterTestHelpers.networkQueryId3,AdapterTestHelpers.networkQueryId2, AdapterTestHelpers.networkQueryId1))(queryMasterIds)
  }

  @Test
  def testReadI2b2AdminPreviousQueriesAscending():Unit = {
    val requestInstance = requestPreviousQueries.copy(sortOrder = SortOrder.Ascending)
    val request: Request[IO] = Request(method = Method.POST, uri = Uri.unsafeFromString(s"/xml")).withEntity(requestInstance.toI2b2String)

    val response: Response[IO] = extractResponse(request)
    assertResult(Ok)(response.status)

    val rpqResponse:ReadPreviousQueriesResponse = response.as[String].map(ReadPreviousQueriesResponse.fromI2b2).unsafeRunSync()
    val queryMasterIds = rpqResponse.queryMasters.map(_.networkQueryId)
    assertResult(Seq(AdapterTestHelpers.networkQueryId1,AdapterTestHelpers.networkQueryId2,AdapterTestHelpers.networkQueryId3, AdapterTestHelpers.networkQueryId4))(queryMasterIds)
  }

  @Test
  def testReadPreviousQueriesStartsWith():Unit = {
    val requestInstance = requestPreviousQueries.copy(
      searchString = AdapterTestHelpers.queryName3.take(4),
      sortOrder = SortOrder.Ascending,
      searchStrategy = Strategy.Left
    )

    val request: Request[IO] = Request(method = Method.POST, uri = Uri.unsafeFromString(s"/xml")).withEntity(requestInstance.toI2b2String)

    val response: Response[IO] = extractResponse(request)
    assertResult(Ok)(response.status)

    val rpqResponse:ReadPreviousQueriesResponse = response.as[String].map(ReadPreviousQueriesResponse.fromI2b2).unsafeRunSync()
    val queryMasterIds = rpqResponse.queryMasters.map(_.networkQueryId)
    assertResult(Seq(AdapterTestHelpers.networkQueryId3))(queryMasterIds)
  }

  //i2b2 aim3 case 1 - uses Contains
  @Test
  def testReadPreviousQueriesContains():Unit = {
    val requestInstance = requestPreviousQueries.copy(
      searchString = AdapterTestHelpers.queryName3.take(4),
      sortOrder = SortOrder.Ascending,
      searchStrategy = Strategy.Contains
    )

    val request: Request[IO] = Request(method = Method.POST, uri = Uri.unsafeFromString(s"/xml")).withEntity(requestInstance.toI2b2String)

    val response: Response[IO] = extractResponse(request)
    assertResult(Ok)(response.status)

    val rpqResponse:ReadPreviousQueriesResponse = response.as[String].map(ReadPreviousQueriesResponse.fromI2b2).unsafeRunSync()
    val queryMasterIds = rpqResponse.queryMasters.map(_.networkQueryId)
    assertResult(Seq(AdapterTestHelpers.networkQueryId3))(queryMasterIds)
  }

  //i2b2 aim3 case 2 - find flagged queries from a specific user
  @Test
  def testReadPreviousQueriesOnlyFlaggedFromUser(): Unit = {
    val requestInstance = requestPreviousQueries.copy(
      categoryToSearchWithin = Category.Flagged,
      username = UserNameQueryOptions.Exactly(AdapterTestHelpers.authn.username)
    )

    val request: Request[IO] = Request(method = Method.POST, uri = Uri.unsafeFromString(s"/xml")).withEntity(requestInstance.toI2b2String)

    val response: Response[IO] = extractResponse(request)

    assertResult(Ok)(response.status)

    val rpqResponse:ReadPreviousQueriesResponse = response.as[String].map(ReadPreviousQueriesResponse.fromI2b2).unsafeRunSync()
    assertResult(Seq(AdapterTestHelpers.networkQueryId1))(rpqResponse.queryMasters.map(_.networkQueryId))
  }

  val requestPreviousQueries:ReadI2b2AdminPreviousQueriesRequest = {

    net.shrine.aim3.ReadI2b2AdminPreviousQueriesRequest(
      projectId = AdapterTestHelpers.projectId,
      waitTime = AdapterTestHelpers.waitTime,
      authn = AdapterTestHelpers.authn,
      username = UserNameQueryOptions.All,
      searchString = "",
      maxResults = 10,
      startDate = None
    )
  }

  @Before
  def beforeEach(): Unit = {
    AdapterQueryHistoryDb.db.createTables()
    loadTestData()
  }

  @After
  def afterEach(): Unit = {
    AdapterQueryHistoryDb.db.dropTables()
  }

  val queryId = 123
  val localMasterId = "kasjdlsajdklajsdkljasd"
  val bogusQueryId = 999

  val masterId1 = "1"
  val masterId2 = "2"
  val masterId3 = "3"
  val masterId4 = "4"

  val networkQueryId1 = 1L
  val networkQueryId2 = 2L
  val networkQueryId3 = 3L
  val networkQueryId4 = 4L

  val queryName1 = "query-name1"
  val queryName2 = "query-name2"
  val queryName3 = "3query-name2"

  lazy val queryDef1: I2b2QueryDefinition = I2b2QueryDefinition(queryName1, Term("x", "xName"))
  lazy val queryDef2: I2b2QueryDefinition = I2b2QueryDefinition(queryName2, Term("y", "yName"))
  lazy val queryDef3: I2b2QueryDefinition = I2b2QueryDefinition(queryName3, Term("y", "yName"))

  val userId = "some-other-user"

  val domain = "Some-other-domain"
  val anotherDomain = "some-completely-different-domain"

  val password = "some-val"

  lazy val authn: AuthenticationInfo = AuthenticationInfo(domain, userId, Credential(password, isToken = false))
  lazy val authn2: AuthenticationInfo = AuthenticationInfo(authn.domain, "a-different-user", Credential("jkafhkjdhsfjksdhfkjsdg", isToken = false))

  val projectId = "some-project-id"

  import scala.concurrent.duration.DurationInt

  val waitTime: FiniteDuration = 12345.milliseconds

  lazy val queryMaster1: QueryMaster = i2b2Protocol.QueryMaster(masterId1, networkQueryId1, queryName1, userId, domain, XmlDateHelper.now, flagged = Some(true))
  lazy val queryMaster2: QueryMaster = i2b2Protocol.QueryMaster(masterId2, networkQueryId2, queryName2, userId, domain, XmlDateHelper.now, flagged = Some(false))
  lazy val queryMaster4: QueryMaster = i2b2Protocol.QueryMaster(masterId4, networkQueryId4, queryName2, authn2.username, anotherDomain, XmlDateHelper.now, flagged = Some(true))


  def loadTestData(): Unit = {
    AdapterQueryHistoryDb.db.insertQueryHistoryIO(masterId1, networkQueryId1, authn, queryDef1, isFlagged = true, hasBeenRun = true, flagMessage = None).unsafeRunSync()

    //NB: Sleep for a few ms to make sure queries are stored with timestamps that differ.
    //java.sql.Date values have millisecond-level resolution.
    Thread.sleep(5)

    AdapterQueryHistoryDb.db.insertQueryHistoryIO(masterId2, networkQueryId2, authn, queryDef2, isFlagged = false, hasBeenRun = true, flagMessage = None).unsafeRunSync()

    Thread.sleep(5)

    AdapterQueryHistoryDb.db.insertQueryHistoryIO(masterId3, networkQueryId3, authn2, queryDef3, isFlagged = false, hasBeenRun = true, flagMessage = None).unsafeRunSync()

    Thread.sleep(5)

    AdapterQueryHistoryDb.db.insertQueryHistoryIO(masterId4, networkQueryId4, authn2.copy(domain = anotherDomain), queryDef2, isFlagged = true, hasBeenRun = true, flagMessage = None).unsafeRunSync()
  }

}

case class AgreeablePmAuthorizer(pmPoster:Poster) extends PmAuthorizer {
  override def authorize(projectId: String, neededRoles: Set[String], authn: AuthenticationInfo): AuthorizationStatus = {
    PmAuthorized(User("Full Name",authn.username,authn.domain,Credential.safe,Map.empty,Map.empty))
  }
}

object AgreeablePmAuthorizer {
  def create: AgreeablePmAuthorizer = {
    val config: Config = ConfigSource.config
    val shrineConfig: Config = config.getConfig("shrine")

    val pmEndpoint: EndpointConfig = shrineConfig.getConfigured("pmEndpoint", EndpointConfig(_))

    val pmPoster: Poster = Poster(pmEndpoint)
    AgreeablePmAuthorizer(pmPoster)
  }
}


