package net.shrine.api

import cats.effect.{ContextShift, IO, Timer}
import cats.implicits.catsSyntaxApplicative
import net.shrine.adapter.QueuedQueriesPoller
import net.shrine.adapter.mappings.AdapterMappings
import net.shrine.aim3.Aim3Service
import net.shrine.api.ontology.OntologyService
import net.shrine.authentication.AuthenticationType
import net.shrine.authz.AuthzHttp4sService
import net.shrine.config.{ConfigExtensions, ConfigSource}
import net.shrine.http4s.catsio.{ExecutionContexts, RepeatedIOTask, RetryIO}
import net.shrine.http4s.servlet.ShrineServletContextSyntax
import net.shrine.hub.data.client.HubClient
import net.shrine.hub.data.service.HubDataService
import net.shrine.hub.mom.{OkToRetry, ShrineMomClient}
import net.shrine.hub.{HubDispatcher, HubLifecycle, OverdueResultsPoller}
import net.shrine.log.Loggable
import net.shrine.messagequeuemiddleware.MessageQueueWebApi
import net.shrine.protocol.version.v2.{NodeSystemSpec, UpdateNodeSystemSpec}
import net.shrine.qep.QepService
import net.shrine.qep.staticdata.StaticDataService
import net.shrine.receiver.Receiver

import javax.servlet.annotation.WebListener
import javax.servlet.{ServletContextEvent, ServletContextListener}
import scala.annotation.unused
import scala.concurrent.duration.{DurationInt, FiniteDuration}
import scala.language.postfixOps

/**
  * This class is needed to load the http4s services in a servlet container
  * Created by marc-danie on 7/24/18.
  */
@WebListener
class Bootstrap extends ServletContextListener with Loggable {
  implicit val cs: ContextShift[IO] = IO.contextShift(ExecutionContexts.dispatchExecutionContext)

  private def afterInitialized(@unused sce: ServletContextEvent): IO[Unit] = {
    ifIO(ConfigSource.config.getBoolean("shrine.hub.create")) {
      implicit val timer: Timer[IO] = IO.timer(ExecutionContexts.dispatchExecutionContext)
      //todo only really needed for the Legacy queues. Maybe switch off for other cases
      //todo find some way to determine that the hub is ready to receive. Wait for that.
      RetryIO
        .keepTryingBounded(HubLifecycle.queuesFromDatabase(), 5 seconds, 12, {case OkToRetry(_) => true},"Start queues")
        .redeemWith((t: Throwable) => IO(error("Something went wrong while starting the hub's queues", t)), //todo see SHRINE-3663 . This is a stop-shrine problem
          _ => IO(info("Hub's queues started successfully")))
    }.
      flatMap(_ => sendNodeSystemSpec()). //Go ahead and send the first one
      flatMap ( _ => nodeSystemSpecSender.restartIO())
  }

  private def sendNodeSystemSpec(): IO[Unit] = {
    val updateIO: IO[UpdateNodeSystemSpec] = for {
      node <- HubClient.getLocalNodeIO
      //todo other IO results needed for future create method parameters go here
    } yield {
      val nodeSystemSpec = NodeSystemSpec.create(node)
      UpdateNodeSystemSpec(nodeSystemSpec)
    }

    updateIO.flatMap { update: UpdateNodeSystemSpec =>
      ShrineMomClient.sendToHubIO(
        subjectId = update.nodeSystemSpec.id,
        envelopeContents = update,
        envelopeContentsCompanion = UpdateNodeSystemSpec,
        logString = "Update node system spec at hub",
      )
    }.handleErrorWith(x => IO(error("Problem sending nodeSystemSpec", x)))
  }

  lazy val nodeSystemSpecSender: RepeatedIOTask = {
    implicit val timer: Timer[IO] = IO.timer(ExecutionContexts.httpClientExecutionContext)

    RepeatedIOTask.scheduleIndefinitely(
      initialDelay = FiniteDuration(0, "seconds"),
      interval = ConfigSource.config.getFiniteDuration("shrine.updateNodeSystemSpec.interval"),
      task = () => sendNodeSystemSpec(),
      name = "Send NodeSystemSpecs"
    )
  }

//todo clean this up to just one unsafeRunSync after cats upgrade SHRINE2020-1218 (and move afterInitialized below if it survives)
  override def contextInitialized(sce: ServletContextEvent): Unit = {
    //todo why is servlet context name null ?  SHRINE-3044  debug(s"contextInitialized for ${sce.getServletContext.getServletContextName}")
    info(s"contextInitialized for ${sce.getServletContext.getContextPath}")

    RepeatedIOTask.startInParallelAndWaitForAll(
      ifIO(ConfigSource.config.getBoolean("shrine.hub.create")) {initHub(sce)}, // receiver starts after hub is set up
      ifIO(ConfigSource.config.getBoolean("shrine.adapter.create")) {initAdapter()},
    )
      .map(_ => info("Support services started - about to mount web api services"))
      .flatMap(_ => mountServices(sce))
      .flatMap(_ => Receiver.restart())
      .attempt
      .flatMap {
        case Left(t)=> IO(abortStartup(t))
        case Right(_) => IO(info("Shrine startup completed."))
      }
      .unsafeRunSync()

    afterInitialized(sce).unsafeRunAsyncAndForget() // Must be async so tomcat can set up the endpoints
  }

  private def abortStartup(cause: Throwable): Unit = {
    error("Shrine startup aborted", cause)
    throw new RuntimeException("Shrine startup aborted", cause) // This stops the servlet
  }

  private def initHub(sce: ServletContextEvent): IO[Unit] =
    RepeatedIOTask.startInParallelAndWaitForAll(
      IO(info(s"This node is a shrine hub.")),
      mountHubServices(sce),
      HubDispatcher.restart(),
      OverdueResultsPoller.restart()
    )

  private def initAdapter(): IO[Unit] = for {
    _ <- AdapterMappings.compareAndReloadMappings(ConfigSource.config.getString("shrine.adapter.adapterMappingsFileName"))
    _ <- QueuedQueriesPoller.restart()
  } yield ()

  private def mountServices(sce: ServletContextEvent):IO[Unit] = IO {
    //TODO: find out why AutoSlash middleware does not work - SHRINE-2581

    ShrineServletContextSyntax.mountService(
      context = sce.getServletContext,
      service = StaticDataService().service,
      name = "SHRINE Static Data Service",
      mapping = "/staticData/*"
    )

    // SHRINE2020-1331
    // In Bootstrap.scala, should the code for SSO be in its own Servlet?
    if(ConfigSource.config.getString("shrine.queryEntryPoint.authenticationType") == AuthenticationType.Sso.name) {
      ShrineServletContextSyntax.mountService(
        context = sce.getServletContext,
        service = AuthzHttp4sService.service,
        name = "Auth Endpoints Service",
        mapping = "/authorizer/*"
      )
    }

    if(ConfigSource.config.getBoolean("shrine.queryEntryPoint.create")) {
      ShrineServletContextSyntax.mountService(
        context = sce.getServletContext,
        name = "SHRINE Ontology Service",
        service = OntologyService().service,
        mapping = "/ontology/*"
      )

      ShrineServletContextSyntax.mountService(
        context = sce.getServletContext,
        service = QepService().router,
        name = "SHRINE QEP Service",
        mapping = "/qep/*"
      )
    }

    ShrineServletContextSyntax.mountService(
      context = sce.getServletContext,
      service = Aim3Service().service,
      name = "SHRINE Aim3 Service",
      mapping = "/aim3/*"
    )

    ShrineServletContextSyntax.mountService(
      context = sce.getServletContext,
      name = "SHRINE CSP Violation Report Service",
      service = CSPReportService().service,
      mapping = "/csp/*"
    )
  }

  private def mountHubServices(sce: ServletContextEvent):IO[Unit] = IO {
    HubLifecycle.initHubFromConfigIfEmptyIO().unsafeRunSync()

    //todo do not bother if using AWS SQS
    ShrineServletContextSyntax.mountService(
      context = sce.getServletContext,
      service = MessageQueueWebApi().service,
      name = "SHRINE MOM",
      mapping = "/mom/*"
    )
    ShrineServletContextSyntax.mountService(
      context = sce.getServletContext,
      service = HubDataService.service,
      name = "SHRINE Hub Data Service",
      mapping = "/hub/*"
    )

    debug(s"services mounted for ${sce.getServletContext.getServletContextName}")
  }

  override def contextDestroyed(sce: ServletContextEvent): Unit = {
    info("Shrine shutdown has been initiated.")
    RepeatedIOTask.startInParallelAndWaitForAll(
      nodeSystemSpecSender.stopIO(),
      ifIO(ConfigSource.config.getBoolean("shrine.adapter.create")) (destroyAdapter),
      ifIO(ConfigSource.config.getBoolean("shrine.hub.create")) (destroyHub),
      Receiver.stop(),
    ).flatMap(_ => ExecutionContexts.shutdown())
      .unsafeRunSync()
  }

  private def destroyAdapter: IO[Unit] = QueuedQueriesPoller.stop()

  private def destroyHub: IO[Unit] = OverdueResultsPoller.stop() *> HubDispatcher.stop()

  private def ifIO(condition: => Boolean)(io: IO[Unit]): IO[Unit] = io.whenA(condition)
}
