package net.shrine.receiver

import java.util.concurrent.atomic.AtomicReference
import cats.effect.{CancelToken, IO}
import ch.qos.logback.classic.Level
import net.shrine.adapter.RunQueryInterrogator
import net.shrine.adapter.dao.AdapterQueryHistoryDb
import net.shrine.config.ConfigSource
import net.shrine.hub.data.client.HubClient
import net.shrine.hub.mom.ShrineMomClient
import net.shrine.log.Log
import net.shrine.problem.{AbstractProblem, ProblemSources}
import net.shrine.protocol.version.v2.{UpdateResultWithError, Result, ResultStatus, RunQueryForResult, UpdateQueryAtAdapter, UpdateQueryAtAdapterWithFlagging, UpdateQueryAtAdapterWithName, UpdateQueryAtQep, UpdateResult}
import net.shrine.protocol.version.{Envelope, JsonText, QueryId}
import net.shrine.qep.querydb.QepQueryDb

/**
 * Receives messages no matter what, and passes them to the Dispatcher to process
 */

object Receiver {

  private val atomicCancelToken = new AtomicReference[CancelToken[IO]](null)

  def restart(): IO[Unit] = IO {
    synchronized {
      stop()
      atomicCancelToken.set(ShrineMomClient.receiveUntilStop(
        HubClient.getLocalNodeIO.map(_.momQueueName),
        dispatch
      ))
    }
  }

  def stop(): IO[Unit] = IO {
    synchronized {
      Option(atomicCancelToken.getAndSet(null)).foreach { ct: CancelToken[IO] =>
        Log.debug(s"Stopping Dispatcher")
        ct.unsafeRunSync()
        Log.debug(s"Stopped Dispatcher")
      }
    }
  }

  private val runQueryInterrogator: Option[RunQueryInterrogator] = {
    if (ConfigSource.config.getBoolean("shrine.adapter.create")) Some(RunQueryInterrogator(ConfigSource.config))
    else None
  }

  //true to complete, false to not complete
  private def dispatch(envelope: Envelope): IO[Boolean] = {

    Log.debug(s"envelope holds a ${envelope.contentsType}")

    envelope match {
      //Adapter commands
      case Envelope(contentsType, _, contents, _) if contentsType == RunQueryForResult.envelopeType =>
        val runQueryForResult: RunQueryForResult = RunQueryForResult.tryRead(new JsonText(contents)).get
        Log.info(s"Received command to run query ${runQueryForResult.query.id}")
        //if there's an adapter run the query. Otherwise complain.
        runQueryInterrogator.fold {
          val queryId = runQueryForResult.query.id
          val errorUpdate = UpdateResultWithError(runQueryForResult.resultProgress.toError(HasNoAdapter(runQueryForResult.query.id),ResultStatus.ErrorInShrine))
          ShrineMomClient.sendToHubIO(queryId,errorUpdate,UpdateResult,s"$errorUpdate to hub").
            handleErrorWith(t => IO(Log.error(s"Exception while sending misconfigured adapter error update $errorUpdate", t))).
            flatMap(_ => IO(true))
        } { rqa =>
          rqa.startRunQueryForExpectedResult(runQueryForResult).
            flatMap(_ => IO(true))
        }
      case Envelope(contentsType, _, contents, _) if contentsType == UpdateQueryAtAdapter.envelopeType =>
        val updateQuery: UpdateQueryAtAdapter = UpdateQueryAtAdapter.tryRead(new JsonText(contents)).get
        Log.info(s"Received command to update query $updateQuery at the adapter")
        val updateSomething: IO[Unit] = updateQuery match {
          case updateQueryName: UpdateQueryAtAdapterWithName => renameQuery(updateQueryName)
          case updateQueryFlagging: UpdateQueryAtAdapterWithFlagging => setQueryFlag(updateQueryFlagging)
          case _ => throw new IllegalStateException(s"Unexpected UpdateQueryAtAdapter command type ${updateQuery.getClass.getSimpleName} $updateQuery")
        }
        updateSomething.flatMap(_ => IO(true))
      //QEP commands
      case Envelope(contentsType, _, contents, _) if contentsType == UpdateQueryAtQep.envelopeType =>
        val updateQueryAtQep = UpdateQueryAtQep.tryRead(new JsonText(contents)).get
          QepQueryDb.db.updateQepQueryIO(updateQueryAtQep).flatMap(_ => IO(true))
      case Envelope(contentsType, _, contents, _) if contentsType == Result.envelopeType => //todo remove with SHRINE2020-1365
        val result = Result.tryRead(new JsonText(contents)).get
        QepQueryDb.db.insertQueryResultIO(result).flatMap(_ => IO(true))
      case Envelope(contentsType, _, contents, _) if contentsType == UpdateResult.envelopeType =>
        val updateResult = UpdateResult.tryRead(new JsonText(contents)).get
        QepQueryDb.db.insertQueryResultIO(updateResult.result).flatMap(_ => IO(true))
      //something unexpected
      case _ => throw new IllegalArgumentException(s"Unsupported Envelope contents type of ${envelope.contentsType} in ${envelope.briefString}")
    }
  }

  private def renameQuery(updateQueryName: UpdateQueryAtAdapterWithName): IO[Unit] = runQueryInterrogator.map { _ =>
     AdapterQueryHistoryDb.db.renameQueryIO(
      updateQueryName.queryId.underlying,
      updateQueryName.queryName)
  }.getOrElse(IO.unit)

  private def setQueryFlag(updateQueryFlagging: UpdateQueryAtAdapterWithFlagging): IO[Unit] = {
    if (updateQueryFlagging.flagged) runQueryInterrogator.map(_ => AdapterQueryHistoryDb.db.flagQueryIO(
      updateQueryFlagging.queryId.underlying,
      updateQueryFlagging.flaggedMessage)).getOrElse(IO.unit)

    else runQueryInterrogator.map(_ => AdapterQueryHistoryDb.db.unFlagQueryIO(
      updateQueryFlagging.queryId.underlying)).getOrElse(IO.unit)
  }
}

//noinspection ScalaFileName
case class HasNoAdapter(queryId:QueryId) extends AbstractProblem(ProblemSources.Adapter) {
  override def logLevel: Level = Level.WARN
  override val summary: String = "This SHRINE node is not configured to have an adapter."
  override val description = s"Received command to run query $queryId but this node has no shrine adapter."
}