package net.shrine.authentication.http4s

import cats.data.{Kleisli, OptionT}
import cats.effect.IO
import net.shrine.authentication.AuthenticationType
import net.shrine.authentication.pm.User
import net.shrine.config.ConfigSource
import net.shrine.config.ConfigExtensions
import org.http4s.Credentials.Token
import org.http4s.headers.Authorization
import org.http4s.headers.Authorization.HeaderT
import org.http4s.{AuthScheme, AuthedRoutes, HttpRoutes, Response, Status}


object UserAuthentication {
  lazy val domain: String = ConfigSource.config.getOption("shrine.webclient.domain", _.getString).getOrElse(ConfigSource.config.getString("shrine.i2b2Domain"))

  def addUnauthorizedMsg(message: String) =
    Kleisli { resp: Response[IO] =>
      OptionT[IO, Response[IO]] {
        if (resp.status == Status.Unauthorized)
          IO(Some(resp.withEntity(message)))
        else IO(Some(resp))
      }
    }
}

//Allows Either Basic or Bearer Authentication
object AuthMiddlewareSelector {
  def apply(service: AuthedRoutes[User, IO]): HttpRoutes[IO] = Kleisli({ request =>

    val authHeader: Option[HeaderT] = request.headers.get(Authorization)
    val authType: AuthenticationType = ConfigSource.config.get("shrine.queryEntryPoint.authenticationType", {aName =>
      AuthenticationType.namesToAuthenticationTypes.getOrElse(aName, throw new IllegalArgumentException(s"$aName is not a valid authentication type"))
    })

    (authType, authHeader) match {
      case (AuthenticationType.Sso, _) => SSOAuthentication.authMiddleware(service)(request)
      case (AuthenticationType.Pm, Some(Authorization(Token(AuthScheme.Bearer, _)))) => BearerAuthMiddleware(service)(request)
      case (AuthenticationType.Pm,_) => BasicAuthMiddleware(service)(request)
    }
  })
}

object XRequestedWithMiddleware {
  import org.http4s.headers.{`WWW-Authenticate`, `X-Requested-With`}

  def apply(service: HttpRoutes[IO]): HttpRoutes[IO] = Kleisli({ request =>

    val authHeader = request.headers.get(`X-Requested-With`)

    authHeader.find(_.value.equalsIgnoreCase("xmlhttprequest")).map( _ => {
      service (request).map {
        case Status.Unauthorized (resp) => resp.withHeaders (resp.headers.filterNot (header => header.is (`WWW-Authenticate`) ) )
        case resp => resp
      }
    }).getOrElse(service(request))
  })

}

