package net.shrine.authentication.http4s

import cats.data.OptionT
import cats.effect.{ConcurrentEffect, ContextShift, IO}
import cats.implicits.{catsSyntaxEither => _}
import net.shrine.authentication.pm.User
import org.http4s.dsl.impl.{->, /, Auth}
import org.http4s.dsl.io.{GET, Ok, Root, Unauthorized, http4sOkSyntax}
import org.http4s.headers.Authorization
import org.http4s.{AuthedRoutes, BasicCredentials, Header, Headers, HttpRoutes, Method, Request, Response, Uri}
import org.junit.runner.RunWith
import org.scalatest.FlatSpec
import org.scalatestplus.junit.JUnitRunner

import scala.concurrent.ExecutionContext.Implicits.global
/**
  * Test web api of BasicAuthentication
  */
@RunWith(classOf[JUnitRunner])
class BasicAuthenticationTest extends FlatSpec with Auth  {

  //Auth is needed for 'as' to work in 'Root / "private" as user'
  private val authService  =  AuthedRoutes.of[User, IO] {
    case GET -> Root / "private" as user =>
      Ok(s"This page is protected using HTTP authentication; logged in as ${user.username}")
  }
  val wrappedAuthService: HttpRoutes[IO] = BasicAuthMiddleware(authService)

  def extractResponse(request: Request[IO]): Response[IO] = {
    implicit val contextShift: ContextShift[IO] = IO.contextShift(global)
    implicit val concurrentEffect: ConcurrentEffect[IO] = IO.ioConcurrentEffect(contextShift)

    val responseOptionIo: OptionT[IO, Response[IO]] = wrappedAuthService.run(request)

    responseOptionIo.map { r: Response[IO] => r }.fold {
      fail("No response from service")
    } { r: Response[IO] => r
    }.unsafeRunSync()
  }

  "An authorized user" should "be allowed access to the service" in {
    val username = "admin"
    val password = "pwd"
    val pingRequest: Request[IO] = Request(method = Method.GET,
      uri = Uri.unsafeFromString("/private"),
      headers = Headers.of(Authorization(BasicCredentials(username, password)))
    )

    val response: Response[IO]= extractResponse(pingRequest)

    assertResult(Ok)(response.status)
    val entityText = response.as[String].unsafeRunSync()
    assertResult("This page is protected using HTTP authentication; logged in as admin")(entityText)
  }

  "A user with no username or password" should "not be allowed access to the service" in {
    val pingRequest: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString("/private"))

    val response: Response[IO]= extractResponse(pingRequest)

    assertResult(Unauthorized)(response.status)
    val entityText = response.as[String].unsafeRunSync()
    assertResult(BasicAuthentication.unauthorizedMsg)(entityText)
  }

  "A user with a username and no password" should "not be allowed access to the service" in {
    val username = "admin"
    val password = ""
    val pingRequest: Request[IO] = Request(method = Method.GET,
      uri = Uri.unsafeFromString("/private"),
      headers = Headers.of(Authorization(BasicCredentials(username, password))))

    val response: Response[IO]= extractResponse(pingRequest)

    assertResult(Unauthorized)(response.status)
    val entityText = response.as[String].unsafeRunSync()
    assertResult(BasicAuthentication.unauthorizedMsg)(entityText)
  }

  "A user with invalid credentials" should "should not get auth challenge if request is an XMLHttpRequest" in {
    import org.http4s.headers.{`WWW-Authenticate`, `X-Requested-With`}

    val username = "admin"
    val password = ""

    val xmlHttpRequest: Request[IO] = Request(method = Method.GET,
      uri = Uri.unsafeFromString("/private"),
      headers = Headers.of(Authorization(BasicCredentials(username, password)), Header(`X-Requested-With`.name.value, "XMLHttpRequest")))

    val xmlHttpResponse: Response[IO]= extractResponse(xmlHttpRequest)

    assertResult(Unauthorized)(xmlHttpResponse.status)
    assert(!xmlHttpResponse.headers.exists(_.is(`WWW-Authenticate`)))

    val xmlHttpEntityText = xmlHttpResponse.as[String].unsafeRunSync()
    assertResult(BasicAuthentication.unauthorizedMsg)(xmlHttpEntityText)
  }

  "A user with invalid credentials" should "should get auth challenge if request is not an XMLHttpRequest" in {
    import org.http4s.headers.`WWW-Authenticate`

    val username = "admin"
    val password = ""
    val request: Request[IO] = Request(method = Method.GET,
      uri = Uri.unsafeFromString("/private"),
      headers = Headers.of(Authorization(BasicCredentials(username, password))))

    val response: Response[IO]= extractResponse(request)

    assertResult(Unauthorized)(response.status)
    assert(response.headers.exists(_.is(`WWW-Authenticate`)))

    val entityText = response.as[String].unsafeRunSync()
    assertResult(BasicAuthentication.unauthorizedMsg)(entityText)
  }
}
