package net.shrine.http4s.catsio

import java.util.concurrent.atomic.AtomicReference
import cats.data.NonEmptyList
import cats.effect.{ContextShift, IO, Timer}
import fs2.Stream
import fs2.concurrent.SignallingRef
import net.shrine.log.Loggable

import scala.concurrent.duration.FiniteDuration
import scala.util.control.NonFatal

/**
 * Execute IO's periodically. Interface loosely based off Monix's Scheduler.
 */

trait RepeatedIOTask {

  /**
   * Stop all executions of the currently scheduled task immediately.
   */
  def stopIO(): IO[Unit]

  /**
   * Stop all executions of the currently scheduled task immediately, and then rerun
   * the task with the specified parameters.
   */
//todo rename start - and see if it's ever "restarted" instead of just started
  def restartIO(): IO[Unit]
}

object RepeatedIOTask extends Loggable {

  /**
   * Execute all IO's in tasks concurrently using the given ContextShift, and wait for all results, returning Unit.
   */

  def startInParallelAndWaitForAll(tasks: IO[_]*)(implicit cs: ContextShift[IO]): IO[Unit] = {
    info(s"Start of startInParallelAndWaitForAll with ${tasks.size}")
    import cats.implicits.catsSyntaxParallelSequence1
    val result = NonEmptyList.fromList(tasks.toList).fold(IO.unit) {
      _.parSequence *> IO.unit
    }
    info(s"End of startInParallelAndWaitForAll with ${tasks.size}")
    result
  }

  /**
   * Execute task once after the initial delay, and subsequently after interval,
   * until task returns false or the scheduler is stopped by the stop() method.
   */
  def scheduleWithFixedDelay(initialDelay: FiniteDuration,
                             interval: FiniteDuration,
                             task: () => IO[Boolean], //todo return true to keep going?
                             name: String)
                            (implicit timer: Timer[IO], cs: ContextShift[IO]): RepeatedIOTask =
    new RepeatedIOTaskImpl(initialDelay, interval, task, name)

  /**
   * Execute task indefinitely, ignoring output of task, until stop or restart methods are called.
   */
  def scheduleIndefinitely(initialDelay: FiniteDuration,
                           interval: FiniteDuration,
                           task: () => IO[_],
                           name: String)
                          (implicit timer: Timer[IO], cs: ContextShift[IO]): RepeatedIOTask =
    scheduleWithFixedDelay(initialDelay, interval, () => task().flatMap(_ => IO(true)), name)


  /**
   * Scheduler to perform a task repeatedly until the stop condition(s).
   * Will stop when either the stop() method is called, or the task returns false.
   *
   * @param name         The name of the scheduled service.
   * @param initialDelay The time to wait before performing the first execution of task.
   * @param interval     The time to wait between executions of task
   * @param task         The task to be performed. If task should be performed again, return true, else return false.
   * @param timer        The timer to use to schedule the task. Cats effect's beefed up Execution Context
   */
  private class RepeatedIOTaskImpl(initialDelay: FiniteDuration,
                                   interval: FiniteDuration,
                                   task: () => IO[Boolean],
                                   name: String)
                                  (implicit timer: Timer[IO], cs: ContextShift[IO])
    extends RepeatedIOTask {

    private val atomicSwitch: AtomicReference[Option[SignallingRef[IO, Boolean]]] = new AtomicReference(None)

    override def restartIO(): IO[Unit] = {
      val starterIO: IO[Unit] = for{
        _ <- stopIO()
        _ <- IO(info(s"Started $name"))
        _ <- repeatWithSignal()
      } yield ()
      starterIO.start.map(_ => info(s"Ready to start $name"))
    }

    override def stopIO(): IO[Unit] = {
      atomicSwitch.get().map{ switch =>
        for{
          _ <- IO(info(s"About to stop $name"))
          _ <- switch.set(true)
          _ <- IO(info(s"Signalled $name to stop"))
        } yield ()
      }.getOrElse{IO(info(s"$name not running"))}
    }

    private def repeatWithSignal(): IO[Unit] = {
      def wrappedTask(): IO[Unit] = {
        task().flatMap { keepGoing =>
          if (!keepGoing) stopIO()
          else IO.unit
        }.handleErrorWith {
          case NonFatal(x) =>
            error(s"Caught exception in $name", x)
            IO.raiseError(x)
        }
      }

      def repeater(): Stream[IO, Unit] = Stream.repeatEval{
        wrappedTask()
      }.metered(interval) //todo maybe debounce instead

      val switchCreatorIO: IO[SignallingRef[IO, Boolean]] = SignallingRef[IO, Boolean](false).map { s =>
        atomicSwitch.set(Option(s))
        s
      }

      val interruptable: Stream[IO, Unit] = for {
        switch <- Stream.eval(switchCreatorIO)
        _ <- repeater().interruptWhen(switch)
      } yield()

      //todo metered doesn't start until after the initial delay plus its own interval - fix it to respect the initial delay with SHRINE2020-1218
      IO.sleep(initialDelay).flatMap(_ => interruptable.compile.drain).map(_ => info(s"End of $name"))
    }
  }
}
