package net.shrine.http4s.catsio

import cats.effect.IO

import scala.collection.mutable.ListBuffer
import scala.concurrent.duration.FiniteDuration
import cats.effect.laws.util.TestContext
import org.junit.jupiter.api.Assertions.{assertEquals, assertFalse, assertTrue}
import org.junit.jupiter.api.TestInstance.Lifecycle
import org.junit.jupiter.api.{AfterAll, BeforeAll, BeforeEach, Disabled, Test, TestInstance}

@TestInstance(Lifecycle.PER_CLASS)
@Disabled //todo reenable after the cats IO upgrade see SHRINE2020-1439
class IOSchedulerTest {

  @BeforeAll
  def beforeAll():Unit = {
    IOSchedulerTest.before()
  }

  @AfterAll
  def afterAll():Unit = {
    IOSchedulerTest.after()
  }

  val output: ListBuffer[Int] = ListBuffer()

  @BeforeEach
  def resetOut(): Unit = {
     output.clear()
    IOSchedulerTest.out.reset()
  }

  def noDelaySchedulerOnce: TestContext => RepeatedIOTask = IOSchedulerTest.appenderScheduler(
    initialDelay = IOSchedulerTest.noDelay,
    interval = IOSchedulerTest.noDelay,
    limit = 1,
    output = output,
    name = "noDelaySchedulerOnce"
  )

  def noDelaySchedulerMany: TestContext => RepeatedIOTask = IOSchedulerTest.appenderScheduler(
    initialDelay = IOSchedulerTest.noDelay,
    interval = IOSchedulerTest.noDelay,
    limit = 100,
    output = output,
    name = "noDelaySchedulerMany"
  )

  def shortDelaySchedulerOnce: TestContext => RepeatedIOTask = IOSchedulerTest.appenderScheduler(
    initialDelay = IOSchedulerTest.shortDelay,
    interval = IOSchedulerTest.noDelay,
    limit = 1,
    output = output,
    name = "shortDelaySchedulerOnce"
  )

  def longDelaySchedulerOnce: TestContext => RepeatedIOTask = IOSchedulerTest.appenderScheduler(
    initialDelay = IOSchedulerTest.longDelay,
    interval = IOSchedulerTest.noDelay,
    limit = 1,
    output = output,
    name = "longDelaySchedulerOnce"
  )

  def schedulerShortInterval: TestContext => RepeatedIOTask = IOSchedulerTest.appenderScheduler(
    initialDelay = IOSchedulerTest.noDelay,
    interval = IOSchedulerTest.shortDelay,
    limit = 100,
    output = output,
    name = "schedulerShortInterval"
  )

  def schedulerDelayAndInterval: TestContext => RepeatedIOTask = IOSchedulerTest.appenderScheduler(
    initialDelay = IOSchedulerTest.shortDelay,
    interval = IOSchedulerTest.shortDelay,
    limit = 2,
    output = output,
    name = "schedulerDelayAndInterval"
  )

  def stackBlowingTask: TestContext => RepeatedIOTask = IOSchedulerTest.appenderScheduler(
    initialDelay = IOSchedulerTest.noDelay,
    interval = IOSchedulerTest.noDelay,
    limit = 1000000,
    output = output,
    name = "stackBlowingTask"
  )

  @Test
  def testNoDelayOnce(): Unit = IOSchedulerTest.testAsync { ec: TestContext =>
    val scheduler = noDelaySchedulerOnce(ec)
    assertEquals(0, output.length)
    scheduler.restartIO().unsafeRunSync()
    assertEquals(0, output.length)
    ec.tick()
    assertEquals(1, output.length)
    ec.tick()
    assertEquals(1, output.length)
  }

  @Test
  def testNoDelayScheduleMany(): Unit = IOSchedulerTest.testAsync { ec: TestContext =>
    val scheduler = noDelaySchedulerMany(ec)
    scheduler.restartIO().unsafeRunSync()

    assertEquals(0, output.length)
    ec.tick(IOSchedulerTest.noDelay)
    // since no delay interval, all done at once.
    assertEquals(100, output.length)
  }

  @Test
  def testShortDelayOnce(): Unit = IOSchedulerTest.testAsync { ec: TestContext =>
    val scheduler = shortDelaySchedulerOnce(ec)
    scheduler.restartIO().unsafeRunSync()

    ec.tick(IOSchedulerTest.noDelay)
    assertEquals(0, output.length)
    ec.tick(IOSchedulerTest.shortDelay)
    assertEquals(1, output.length)
    ec.tick() // all done already
    assertEquals(1, output.length)
  }

  @Test
  def testLongDelayOnce(): Unit = IOSchedulerTest.testAsync { ec: TestContext =>
    val scheduler = longDelaySchedulerOnce(ec)
    scheduler.restartIO().unsafeRunSync()

    ec.tick(IOSchedulerTest.noDelay)
    assertEquals(0, output.length)
    ec.tick(IOSchedulerTest.shortDelay)
    assertEquals(0, output.length)
    ec.tick(IOSchedulerTest.longDelay)
    assertEquals(1, output.length)
  }

  @Test
  def testShortInterval(): Unit = IOSchedulerTest.testAsync { ec: TestContext =>
    val scheduler = schedulerShortInterval(ec)
    scheduler.restartIO().unsafeRunSync()

    ec.tick(IOSchedulerTest.noDelay)
    assertEquals(1, output.length)
    ec.tick(IOSchedulerTest.noDelay)  // must wait for interval
    assertEquals(1, output.length)
    ec.tick(IOSchedulerTest.shortDelay)
    assertEquals(2, output.length)
  }

  @Test
  def testDelayAndInterval(): Unit = IOSchedulerTest.testAsync { ec: TestContext =>
    val scheduler = schedulerDelayAndInterval(ec)
    scheduler.restartIO().unsafeRunSync()

    assertEquals(0, output.length)
    ec.tick(IOSchedulerTest.noDelay)
    assertEquals(0, output.length)

    ec.tick(IOSchedulerTest.shortDelay)
    assertEquals(1, output.length)

    assertEquals(1, output.length)
    ec.tick(IOSchedulerTest.noDelay)
    assertEquals(1, output.length)

    ec.tick(IOSchedulerTest.shortDelay)
    assertEquals(2, output.length)
  }

  @Test
  def testStackSafety(): Unit = IOSchedulerTest.testAsync { ec: TestContext =>
    val scheduler = stackBlowingTask(ec)
    assertEquals(0, output.length)
    scheduler.restartIO().unsafeRunSync()
    assertEquals(0, output.length)
    // If the implementation wasn't stack safe this would blow up
    ec.tick()
    assertEquals(1000000, output.length)
  }

  @Test
  def testLoggedMessages(): Unit = IOSchedulerTest.testAsync { ec: TestContext =>
    val scheduler = schedulerDelayAndInterval(ec)
    assertEquals("", IOSchedulerTest.out.toString)
    scheduler.restartIO()
    assertEquals("", IOSchedulerTest.out.toString)
    // actually have to run to begin anything
    scheduler.restartIO().unsafeRunSync()
    assertTrue(IOSchedulerTest.out.toString.contains("Started schedulerDelayAndInterval"))
    assertEquals(0, output.length) // Begun, but waiting on timer

    ec.tick(IOSchedulerTest.shortDelay)
    assertFalse(IOSchedulerTest.out.toString.contains("End of schedulerDelayAndInterval"))
    assertEquals(1, output.length)

    ec.tick(IOSchedulerTest.shortDelay)
    assertTrue(IOSchedulerTest.out.toString.contains("End of schedulerDelayAndInterval"))
    assertEquals(2, output.length)
  }

  @Test
  def testStop(): Unit = IOSchedulerTest.testAsync { ec: TestContext =>
    val scheduler = schedulerDelayAndInterval(ec)
    scheduler.restartIO().unsafeRunSync()

    ec.tick(IOSchedulerTest.shortDelay)
    assertEquals(List(0), output.toList)

    scheduler.stopIO().unsafeRunSync()
    assertEquals(List(0), output.toList)
    assertEquals(true, IOSchedulerTest.out.toString.contains("Stopping schedulerDelayAndInterval"))
    assertEquals(true, IOSchedulerTest.out.toString.contains("Stopped schedulerDelayAndInterval"))
  }

  @Test
  def testRestart(): Unit = IOSchedulerTest.testAsync { ec: TestContext =>
    val scheduler = schedulerDelayAndInterval(ec)
    scheduler.restartIO().unsafeRunSync()

    ec.tick(IOSchedulerTest.shortDelay)
    assertEquals(List(0), output.toList)

    ec.tick(IOSchedulerTest.shortDelay)
    assertEquals(List(0, 1), output.toList)

    scheduler.restartIO().unsafeRunSync()
    ec.tick(IOSchedulerTest.shortDelay)
    assertEquals(List(0, 1, 2), output.toList) // Blew past original limit

    assertEquals(2, IOSchedulerTest.out.toString.split("\n").count(_.contains("Started schedulerDelayAndInterval")))
    assertEquals(2, IOSchedulerTest.out.toString.split("\n").count(_.contains("End of schedulerDelayAndInterval.pollUntilStop()")))
  }
}

object IOSchedulerTest extends LoggingTest {
  private val noDelay: FiniteDuration = FiniteDuration(0, "milliseconds")
  private val shortDelay: FiniteDuration = FiniteDuration(500, "milliseconds")
  private val longDelay: FiniteDuration = FiniteDuration(3, "seconds")

  // Simulate some side-effecting operation
  def appendTask(limit: Integer): ListBuffer[Int] => () => IO[Boolean] = {
    var elem = 0
    l: ListBuffer[Int] => () => IO {
      l += elem
      elem = elem + 1
      if (elem < limit) true else false
    }
  }

  def testAsync(f: TestContext => Unit): Unit = f(TestContext())

  def appenderScheduler(initialDelay: FiniteDuration,
                        interval: FiniteDuration,
                        limit: Int, output:
                        ListBuffer[Int],
                        name: String = "IOSchedulerTest"):
  TestContext => RepeatedIOTask = { ec: TestContext =>
    RepeatedIOTask.scheduleWithFixedDelay(
      name = name,
      initialDelay = initialDelay,
      interval = interval,
      task = appendTask(limit)(output)
    )(ec.timer[IO],ec.contextShift)
  }
}