package net.shrine.http4s.servlet

import cats.data.OptionT
import cats.effect.{ConcurrentEffect, ContextShift, IO}
import org.http4s.{Header, HttpRoutes, Method, Request, Response, Uri}
import org.http4s.dsl.impl.->
import org.http4s.dsl.io.{GET, NotFound, Ok, Root, http4sNotFoundSyntax, http4sOkSyntax}
import org.http4s.headers.`Strict-Transport-Security`
import org.junit.jupiter.api.Assertions.{assertEquals, assertTrue, fail}
import org.junit.Test

import scala.concurrent.ExecutionContext.Implicits.global
import scala.concurrent.duration.DurationInt

class ShrineServletContextSyntaxTest{

  private[servlet] def extractResponse(request: Request[IO], service: HttpRoutes[IO]): Response[IO] = {
    implicit val contextShift: ContextShift[IO] = IO.contextShift(global)
    implicit val concurrentEffect: ConcurrentEffect[IO] = IO.ioConcurrentEffect(contextShift)

    val responseOptionIo: OptionT[IO, Response[IO]] = service.run(request)
    responseOptionIo.map { r: Response[IO] => r }.fold {
      fail("No response from service")
    } { r: Response[IO] => r
    }.unsafeRunSync()
  }

  val pingService: HttpRoutes[IO] = HttpRoutes.of[IO] {
    case GET -> Root => Ok("pong")
    case x => NotFound(s"The service does not respond to $x")
  }

  @Test
  def testHSTSResponseHeader(): Unit = {
    implicit val cs: ContextShift[IO] = IO.contextShift(global)

    val shrineService = ShrineServletContextSyntax.wrapResponseHeaders(pingService)
    val pingRequest: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString(s"/ping"))
    val response: Response[IO] = extractResponse(pingRequest, shrineService)

    val hstsHeader: Header = `Strict-Transport-Security`.unsafeFromDuration(
      maxAge = 365.days,
      includeSubDomains = true,
      preload = false
    )

    val responseHstsHeader = response.headers.find(_.equals(hstsHeader))
    responseHstsHeader.fold(fail("Missing HSTS header"))(h => {
      assertTrue(h.equals(hstsHeader))
      assertEquals("max-age=31536000; includeSubDomains", h.value)
    })
  }
}
