package net.shrine.hub.metrics

import com.opencsv.CSVWriter
import net.shrine.hub.data.store.HubDb
import net.shrine.protocol.version.v2.{CountResult, ErrorResult, Node, Result, ResultStatus}
import net.shrine.protocol.version.{DateStamp, NodeId, NodeKey, NodeName, QueryId, ResultId, UserName}

import java.io.File

object ResultCsv {

  def writeResultsCsv(file:File,startDate:DateStamp,endDate:DateStamp,researcher:Option[UserName]):Unit = CommonMaps.writeCsv (file,{ csvWriter: CSVWriter =>
    csvWriter.writeNext(ResultRow.header)

    val queryIds: Seq[QueryId] = researcher.fold(
      HubDb.db.selectQueryIdsIO(startDate, endDate)
    ) { r =>
      val researcherId = CommonMaps.researcherUserNameToResearchers(r).map(_.id)
      HubDb.db.selectQueryIdsForResearcherIdsIO(startDate, endDate, researcherId)
    }.unsafeRunSync()

    queryIds.foreach{queryId =>
      val queryHistory = HubDb.db.selectQueryHistoryIO(queryId).unsafeRunSync()
      val latestResults = HubDb.db.selectMostRecentResultsForQueryIO(queryId).unsafeRunSync()
      val resultHistories = HubDb.db.selectResultHistoryIO(queryId).unsafeRunSync()

      val queryRow = QueryRow(queryHistory,latestResults)
      val resultHistoriesByNode: Seq[(NodeId, Seq[Result])] = resultHistories.groupBy(_.adapterNodeId).
        toSeq.sortBy { r =>
        CommonMaps.nodeIdToNode(r._2.head.adapterNodeId).name.underlying
      }

      val resultRows = resultHistoriesByNode.map { nodeIdToResults: (NodeId, Seq[Result]) =>
        ResultRow(queryRow, nodeIdToResults._2)
      }
      resultRows.foreach { resultRow => csvWriter.writeNext(resultRow.resultArray) }
    }
  })
}

case class ResultRow(
                      id:ResultId,
                      queryRow: QueryRow,
                      adapterNodeShortName:NodeKey,
                      adapterNodeName:NodeName,
                      status:ResultStatus,
                      count:Int,
                      errorCodec: String,
                      errorSummary: String,
                      createDate:DateStamp,
                      resultStatesToTimeInState: Map[ResultStatus, Long]
                    ) {

  def resultArray: Array[String] = {

    val resultArray:Array[String] = Array[String](
      queryRow.id.excelSafeString,
      queryRow.queryName,
      queryRow.researcherUserName.underlying,
      queryRow.researcherDomainName.underlying,
      id.excelSafeString,
      adapterNodeShortName.underlying,
      adapterNodeName.underlying,
      status.statusName,
      count.toString,
      errorCodec,
      errorSummary,
      createDate.toDateString
    ) ++ {
      ResultRow.statusOrder.map {
        resultStatesToTimeInState.get(_).fold("")(_.toString)
      }
    }
    resultArray
  }
}

object ResultRow {

  val ignoreStatuses: Set[ResultStatus] = Set(
    ResultStatus.QueuedByAdapter, //todo not yet reported to the hub
    ResultStatus.QueuedForManualSubmission, //todo if we ever support this feature add this state back
    ResultStatus.UnknownWhileQueuedByCRC,
    ResultStatus.UnknownInTransit,
    ResultStatus.UnknownFinal,
    ResultStatus.ReadyToSubmit, //todo not yet recorded at the hub
  ) ++ ResultStatus.statuses.filter(_.isFinal)

  val statusOrder: Seq[ResultStatus] = ResultStatus.statuses.filterNot(ignoreStatuses.contains)

  val header: Array[String] = {

    val fixedNames = Array(
      "Query Id",
      "Query Name",
      "Researcher User Name",
      "Researcher Authentication Domain",
      "Result Id",
      "Adapter Node Key",
      "Adapter Node Name",
      "Status",
      "Count",
      "Error Codec",
      "Error Summary",
      "Create Date"
    )

    val stateNames = statusOrder.map(_.statusName).map(x => s"Time in $x (ms)")
    fixedNames ++ stateNames
  }

  def apply(
             queryRow: QueryRow,
             resultHistory: Seq[Result]
           ): ResultRow = {
    val latestResult = resultHistory.head
    val node: Node = CommonMaps.nodeIdToNode(latestResult.adapterNodeId)
    val resultStatesAndTimes = resultHistory.map { r: Result => r.status -> r.versionInfo.changeDate }.sortBy(_._2.underlying)
    val zippedResultStatesAndDateStamps: Seq[((ResultStatus, DateStamp), (ResultStatus, DateStamp))] =
      resultStatesAndTimes.zip(resultStatesAndTimes.tail)
    val resultStatesToTimeInState: Map[ResultStatus, Long] = zippedResultStatesAndDateStamps.map{ zippedRsDs =>
      zippedRsDs._1._1 -> (zippedRsDs._2._2.underlying - zippedRsDs._1._2.underlying)
    }.toMap

    ResultRow(
      id = latestResult.id,
      queryRow = queryRow,
      adapterNodeShortName = node.key,
      adapterNodeName = node.name,
      status = latestResult.status,
      count = latestResult match {
        case countResult: CountResult =>
          if(countResult.count != -1) countResult.count
          else 0
        case _ => 0
      },
      errorCodec = latestResult match {
        case errorResult: ErrorResult => errorResult.problemDigest.codec
        case _ => ""
      },
      errorSummary = latestResult match {
        case errorResult: ErrorResult => errorResult.problemDigest.summary
        case _ => ""
      },
      createDate = latestResult.versionInfo.createDate,
      resultStatesToTimeInState = resultStatesToTimeInState,
    )
  }
}