package net.shrine.hub

import cats.effect.{ContextShift, IO, Timer}
import ch.qos.logback.classic.Level
import net.shrine.config.{ConfigExtensions, ConfigSource}
import net.shrine.http4s.catsio.{ExecutionContexts, RepeatedIOTask}
import net.shrine.hub.data.store.HubDb
import net.shrine.log.Loggable
import net.shrine.problem.{AbstractProblem, ProblemSources}
import net.shrine.protocol.version.v2.{Node, Result, ResultStatus, UpdateResultWithError}
import net.shrine.protocol.version.{DateStamp, NodeId, QueryId}

import scala.concurrent.duration.FiniteDuration

object OverdueResultsPoller extends Loggable {

  private val resultsOverdueAfter: FiniteDuration = ConfigSource.config.getFiniteDuration("shrine.hub.result.resultsOverdueAfter")

  private val overdueResultsCheckPeriod: FiniteDuration = ConfigSource.config.getFiniteDuration("shrine.hub.result.overdueResultsCheckPeriod")

  private val overdueResultsBatchSize:Int = ConfigSource.config.getInt("shrine.hub.result.overdueResultsBatchSize")

  implicit val timer: Timer[IO] = IO.timer(ExecutionContexts.crcExecutionContext)
  implicit val concurrentEffect: ContextShift[IO] = IO.contextShift(ExecutionContexts.crcExecutionContext)

  private val scheduler = RepeatedIOTask.scheduleIndefinitely(
    initialDelay = overdueResultsCheckPeriod,
    interval = overdueResultsCheckPeriod,
    task = () => checkForOverdueResults(),
    name = "OldResultsPoller"
  )

  //call this when a query gets queued
  def restart(): IO[Unit] = scheduler.restartIO()

  def stop(): IO[Unit] = scheduler.stopIO()

  def checkForOverdueResults():IO[Unit] = {
    val before: DateStamp = new DateStamp(DateStamp.now.underlying - resultsOverdueAfter.toMillis)
    info(s"Checking for overdue results - results before $before")
    for {
      nodes <- HubDb.db.selectLatestNodesIO
      resultCount <- HubDb.db.countDistinctResultIds(before)
      _ <- IO(info(s"$resultCount results to examine"))
      updateCount <- batchProcessResults(before,resultCount,nodes.map{ node => node.get.id -> node.get}.toMap)
    } yield IO(info(s"Updated $updateCount results"))
  }

  private def batchProcessResults(before:DateStamp, count:Int, nodeIdsToNodes:Map[NodeId, Node]):IO[Int] = {
    import cats.implicits._

    val batchSize = overdueResultsBatchSize
    val batches = (0 to count/batchSize).map(IO(_)).map{ batchIO:IO[Int] =>
      for {
        batch <- batchIO
        overdueResults <- HubDb.db.selectOverdueBatchIO(before, batch, batchSize)
        _ <- IO(debug(s"batch $batch found ${overdueResults.size} overdue results to update"))
        updateCounts <- overdueResults.map(updateAsOverdue(_, nodeIdsToNodes)).sequence
      } yield updateCounts.sum
    }
    batches.toList.sequence.flatMap((updateCounts: Seq[Int]) => IO(updateCounts.sum))
  }

  private def updateAsOverdue(result:Result, nodeIdsToNodes:Map[NodeId, Node]):IO[Int] = {
    debug(s"$result from ${nodeIdsToNodes(result.adapterNodeId).key.underlying} is overdue")
    val problem = QueryAttemptTimeToLiveExceeded(result.queryId,nodeIdsToNodes(result.adapterNodeId),resultsOverdueAfter)
    val overdueResult = result.toError(
      problem = problem,
      status = ResultStatus.ErrorFromCrc,
      statusMessage = Option(problem.summary),
      crcQueryInstanceId = result.crcQueryInstanceId,
      adapterTime = DateStamp.now
    )
    val overdueUpdate = UpdateResultWithError(overdueResult)

    debug(s"Will update $overdueUpdate")
    HubDispatcher.updateResultIO(overdueUpdate).flatMap(_ => IO(1))
  }

}

case class QueryAttemptTimeToLiveExceeded(queryId: QueryId, node:Node, timeToLive:FiniteDuration) extends AbstractProblem(ProblemSources.Hub) {
  override def logLevel: Level = Level.DEBUG
  override val throwable: Option[Throwable] = None
  override val summary: String = s"Shrine's hub has given up on getting a result from ${node.name.underlying} after $timeToLive"
  override val description: String = s"Shrine's hub has given up on getting a result from ${node.name.underlying} for $queryId after $timeToLive"
  override def detailsText: None.type = None
}
