package net.shrine.hub

import cats.effect.IO
import net.shrine.hub.data.store.HubDb
import net.shrine.protocol.version.v2.querydefinition.QueryDefinition
import net.shrine.protocol.version.v2.{ErrorResult, Network, Node, Query, Result, ResultStatus}
import net.shrine.protocol.version.{DateStamp, MomQueueName, ResearcherId}
import org.junit.{After, Before, Test}
import org.scalatestplus.junit.AssertionsForJUnit

class OverdueResultsPollerTest extends AssertionsForJUnit {

  val sixDays: Int = 6 * 24 * 60 * 60 * 1000

  @Test
  def testGiveUpOnSomeResults():Unit = {
    val nodes = (1 to 3).map(i =>
      Node.create(
        name = s"node$i",
        key = s"node$i",
        userDomainName = s"node$i",
        adminEmail = s"admin@node$i.co",
        momId = s"node$i"
    ))

    val query = Query.create(
      queryDefinition = QueryDefinition.standIn,
      breakdownNames = Seq.empty,
      queryName = "expiring query",
      nodeOfOriginId = nodes.head.id,
      researcherId = new ResearcherId(-1),
    )

    val results = nodes.map{ node =>
      Result.create(query,node)
    }.map{r =>
      r.copy(versionInfo = r.versionInfo.copy(changeDate = new DateStamp(r.versionInfo.changeDate.underlying - sixDays)))
    }

    import cats.implicits._
    val setUpDbIo: IO[List[Option[Result]]] = nodes.map(node => HubDb.db.upsertNodeIO(node)).toList.sequence.
      flatMap{_ => results.map(result => HubDb.db.upsertResultIO(result)).toList.sequence}

    setUpDbIo.unsafeRunSync()

    assertResult(3)(HubDb.db.selectOverdueBatchIO(DateStamp.now,0).unsafeRunSync().size)

    OverdueResultsPoller.checkForOverdueResults().unsafeRunSync()

    val actual = HubDb.db.selectMostRecentResultsForQueryIO(query.id).unsafeRunSync()
    assertResult(results.size)(actual.size)

    actual.foreach{ r =>
      assertResult(ResultStatus.ErrorFromCrc)(r.status)
      r match {
        case errorResult:ErrorResult =>
          assertResult(
            QueryAttemptTimeToLiveExceeded.getClass.getName.dropRight(1)
          )(errorResult.problemDigest.codec)
        case _ => fail(s"Unexpected type ${r.getClass} for $r")
      }
    }
    assert(HubDb.db.selectOverdueBatchIO(DateStamp.now,0).unsafeRunSync().isEmpty)
  }

  @Before
  def beforeEach(): Unit = {
    HubDb.db.createTables()

    val expectedNetwork = Network(
      networkName = "testNetwork",
      hubQueueName = MomQueueName("testHub"),
      adminEmail = "yourname@example.com",
      momId = "testNetwork",
      awsSqsConfig = None,
      kafkaConfig = None
    )
    HubDb.db.upsertNetworkIO(expectedNetwork).unsafeRunSync()
  }

  @After
  def afterEach(): Unit = {
    HubDb.db.dropTables()
  }

}
