import React, { useState, useEffect } from "react";
import PropTypes from "prop-types";

import {
  Dialog,
  Button,
  InputLabel,
  Typography,
  Backdrop,
} from "@material-ui/core";

import { Flag } from "models";
import { ValidatedTextField } from "components";
import "./FlagDialog.scss";

export const FlagDialog = ({
  open,
  onClose,
  currentFlag,
  queryId,
  flaggingTexts,
}) => {
  const flagged = currentFlag && currentFlag.flagged;

  const defaultFlaggingInstructions = flagged
    ? "Unflag this query or update the message associated with the flag."
    : "Enter a message for why this query is being flagged.";
  const flaggingInstructions =
    flaggingTexts.queryFlaggingInstructions || defaultFlaggingInstructions;
  const flaggingPlaceholderText =
    flaggingTexts.flaggingPlaceholderText || "Enter a flagging message...";

  const [canSubmit, setCanSubmit] = useState(false);
  const [flagMessage, setFlagMessage] = useState(
    currentFlag ? currentFlag.flagMessage : ""
  );

  const resetFlagMessage = () => {
    if (currentFlag) {
      setFlagMessage(currentFlag.flagMessage);
    } else {
      setFlagMessage("");
    }
  };

  useEffect(() => {
    resetFlagMessage();
  }, [currentFlag]);

  const handleTextChange = (message, valid) => {
    setFlagMessage(message);
    setCanSubmit(valid && !(flagged && message === currentFlag.flagMessage));
  };

  const handleCancel = () => {
    resetFlagMessage();
    onClose(false);
  };
  const handleUnflag = () => {
    setFlagMessage("");
    onClose(true, false, null);
  };
  const handleFlag = () => {
    if (canSubmit) {
      onClose(true, true, flagMessage);
    }
  };

  return (
    <Dialog
      aria-labelledby="spring-modal-title"
      aria-describedby="spring-modal-description"
      className="FlagDialog"
      open={open}
      onClose={onClose}
      closeAfterTransition
      BackdropComponent={Backdrop}
      BackdropProps={{
        timeout: 500,
      }}
    >
      <div className="DialogContent">
        <span className="close-button" onClick={handleCancel} role="button">
          <i className="fa fa-times" aria-hidden="true" />
        </span>
        <div className="title-label-wrapper">
          <Typography className="title">Flag Query</Typography>
        </div>
        <div className="name-label-wrapper">
          <InputLabel>{flaggingInstructions}</InputLabel>
        </div>
        <div className="queryId-label">
          <InputLabel>{`Query ID: ${queryId}`}</InputLabel>
        </div>
        <div>
          <ValidatedTextField
            label="Flag Message"
            onTextChange={handleTextChange}
            multiline
            maxTextLength={1000}
            // we want to allow all characters so do not set invalidCharsRegex
            // invalidCharsRegex="[^ !-~\n]"
            // Note: \n characters are allowed by virtue of the "multiline" property above
            // TODO-XH : do we want to reuse description-input here, or have its own class name?
            className="description-input"
            autoFocus
            placeholder={flaggingPlaceholderText}
            fullWidth
            value={flagMessage}
          />
        </div>
        <div className="button-row">
          {flagged && (
            <Button className="btn-primary" onClick={handleUnflag}>
              Unflag
            </Button>
          )}
          <Button
            onClick={handleFlag}
            className="btn-primary"
            disabled={!canSubmit}
          >
            {flagged ? "Update" : "Flag"}
          </Button>
        </div>
      </div>
    </Dialog>
  );
};

FlagDialog.defaultProps = {
  currentFlag: null,
};

FlagDialog.propTypes = {
  open: PropTypes.bool.isRequired,
  onClose: PropTypes.func.isRequired,
  currentFlag: PropTypes.shape(Flag.propTypes),
  queryId: PropTypes.oneOfType([PropTypes.number, PropTypes.string]).isRequired,
  flaggingTexts: PropTypes.shape({
    queryFlaggingInstructions: PropTypes.string,
    flaggingPlaceholderText: PropTypes.string,
  }).isRequired,
};
