package net.shrine.qep.metrics

import com.opencsv.CSVWriter
import net.shrine.audit
import net.shrine.audit.LongQueryId
import net.shrine.qep.querydb.{QepQuery, QepQueryDb}
import net.shrine.protocol.version.{DateStamp, NodeKey, NodeName, ResearcherId, UserDomainName, UserName}
import net.shrine.protocol.version.v2.{Node, Researcher}

import java.io.File

object ResearcherCsv {

  def writeResearcherCsv(file:File,startDate:DateStamp,endDate:DateStamp,researcher:Option[UserName]):Unit = CommonMaps.writeCsv (file, { csvWriter: CSVWriter =>
    csvWriter.writeNext(ResearcherRow.header)

    val queriesHistory: Seq[QepQuery] = QepQueryDb.db.selectPreviousQueriesHistoryIO(startDate,endDate,researcher.map(_.underlying)).unsafeRunSync()

    val researcherUserNamesToQueriesHistory: Map[audit.UserName, Map[LongQueryId, Seq[QepQuery]]] = queriesHistory.groupBy(_.userName).map{ userNameAndQueries: (audit.UserName, Seq[QepQuery]) =>
      val queryHistories: Map[LongQueryId, Seq[QepQuery]] = userNameAndQueries._2.groupBy(_.networkId).map{ queryIdToQueryStates: (LongQueryId, Seq[QepQuery]) =>
        queryIdToQueryStates._1 -> queryIdToQueryStates._2.sortBy(_.v2Query.versionInfo.itemVersion.underlying)
      }
      userNameAndQueries._1 -> queryHistories
    }

    //for each researcher id - make a row
    val researcherNameToQueryCountsAndDates = researcherUserNamesToQueriesHistory.toSeq.sortBy(_._1).map{x =>
      (
        x._1,
        x._2.size,
        x._2.minBy(_._2.head.v2Query.versionInfo.createDate.underlying)._2.head.v2Query.versionInfo.createDate,
        x._2.maxBy(_._2.head.v2Query.versionInfo.createDate.underlying)._2.head.v2Query.versionInfo.createDate,
      )
    }

    researcherNameToQueryCountsAndDates.map { x =>
      val firstQuery: QepQuery = researcherUserNamesToQueriesHistory(x._1).head._2.head
      val researcher = Researcher(
        id = firstQuery.v2Query.researcherId,
        userName = new UserName(x._1),
        userDomainName = new UserDomainName(firstQuery.userDomain),
        nodeId = firstQuery.v2Query.nodeOfOriginId
      )

      ResearcherRow(
        id = researcher.id,
        userName = researcher.userName,
        userDomainName = researcher.userDomainName,
        createdDate = researcher.versionInfo.createDate, //todo this is going to always be today
        originShortName = new NodeKey("unknown"),
        originName = new NodeName("unknown"),
        queriesStarted = x._2,
        earliestQuery = x._3,
        latestQuery = x._4,
      )
    }.foreach(r => csvWriter.writeNext(r.rowArray))
  })
}

case class ResearcherRow(
                          id:ResearcherId,
                          userName: UserName,
                          userDomainName: UserDomainName,
                          createdDate:DateStamp,
                          originShortName:NodeKey,
                          originName:NodeName,
                          queriesStarted:Int,
                          earliestQuery:DateStamp,
                          latestQuery:DateStamp,
                   ) {
  def rowArray: Array[String] = Array[String](
    id.excelSafeString,
    userName.underlying,
    userDomainName.underlying,
    //Bring back with SHRINE2020-1104    createdDate.toDateString,
    originShortName.underlying,
    originName.underlying,
    queriesStarted.toString,
    earliestQuery.toDateString,
    latestQuery.toDateString,
  )
}

object ResearcherRow{

  val header: Array[String] = Array(
      "Researcher Id",
      "Researcher User Name",
      "Researcher Authentication Domain",
//Bring back with SHRINE2020-1104      "First Query Date",
      "Origin Node Key",
      "Origin Node Name",
      "Number of Queries Started",
      "Earliest Query",
      "Latest Query"
    )

  def apply(
             researcher: Researcher,
             node:Node,
             queriesStarted:Int,
             earliestQuery: DateStamp,
             latestQuery: DateStamp
           ): ResearcherRow = {

    ResearcherRow(
      id =  researcher.id,
      userName = researcher.userName,
      userDomainName = researcher.userDomainName,
      createdDate = researcher.versionInfo.createDate,
      originShortName = node.key,
      originName = node.name,
      queriesStarted = queriesStarted,
      earliestQuery = earliestQuery,
      latestQuery = latestQuery,
    )
  }
}
