package net.shrine.qep.querydb

import java.util.concurrent.atomic.AtomicInteger

import cats.effect.concurrent.Deferred
import cats.effect.{ContextShift, IO, Timer}
import net.shrine.audit.LongQueryId
import net.shrine.http4s.catsio.ExecutionContexts
import net.shrine.log.Log
import org.http4s.Response

import scala.collection.concurrent.{TrieMap, Map => ConcurrentMap}
import scala.concurrent.duration.FiniteDuration

/**
  * A generic appliance for notifying anything watching of changes to a particular query based on id.
  */
object QepQueryDbChangeNotifier {

  type A = Response[IO]

  //todo switch from NetworkQueryIds to Versioned Ids , and from tasks to IOs when this moves to the hub
  case class Trigger(requestId: Long, queryId: LongQueryId, deferred: Deferred[IO,String])

  private val requestIdsToTriggers:ConcurrentMap[Long,Trigger] = TrieMap.empty

  private val requestIdSource: AtomicInteger = new AtomicInteger(0)

  def setTriggerForChange(queryId:LongQueryId, timeout: FiniteDuration, task:IO[A]):IO[A] = {

    val requestId = requestIdSource.getAndIncrement()

    //when a request comes in -
    //1) watch for the change by installing the IO trigger
    val takeAndDoTask:IO[A] = {
      val triggerIO: IO[Trigger] = IO(requestIdsToTriggers.remove(requestId)).map(_.getOrElse(throw new IllegalStateException(s"No action found for $requestId. Should never happen.")))
      triggerIO.flatMap(trigger => IO(Log.debug(s"triggered $trigger"))).flatMap(_ => task)
    }

    val timer: Timer[IO] = IO.timer(ExecutionContexts.longActionExecutionContext)
    implicit val longActionContextShift: ContextShift[IO] = IO.contextShift(ExecutionContexts.longActionExecutionContext)

    val deferredIO: IO[Deferred[IO, String]] = Deferred[IO,String] //todo implicit Concurrent

    deferredIO.flatMap{deferred =>
      val deferredTask: IO[A] = for {
        logSegment <- deferred.get
        _ <- IO(Log.debug(s"Task started $logSegment"))
        a <- takeAndDoTask
      } yield a

      val timeoutTask:IO[A] = IO(Log.debug(s"Task started via timeout")).flatMap(_ => takeAndDoTask)

      val trigger = Trigger(requestId,queryId,deferred)
      requestIdsToTriggers.put(requestId,trigger)
      Log.debug(s"setTriggerForChange requestIdsToTriggers is $requestIdsToTriggers")

      //2) start a timeoutTo to do the task even if it is never triggered
      // if the deferredTask hasn't happened after timeout, do it after waiting long enough
      deferredTask.timeoutTo(timeout,timeoutTask)(timer,longActionContextShift)
    }
  }

  //find all the triggers to trip for queryId, and do them
  def triggerChangesFor(queryId:LongQueryId): IO[Unit] = {
    import cats.implicits._
    Log.debug(s"triggerChangesFor $queryId requestIdsToTriggers is $requestIdsToTriggers")
    requestIdsToTriggers.values.filter(_.queryId == queryId).map{ trigger =>
      Log.debug(s"About to complete $trigger")
      trigger.deferred.complete(s"- triggered for $queryId").flatMap{_ => IO(Log.debug(s"Completed deferred"))}
    }.toList.sequence.flatMap(_ => IO.unit)
    //because the tasks' first step is to remove themselves from the map there's nothing to clean up.
  }
}